// Class CEngine, Version 5.4, Feb 2013

// Version 5.5, Feb-2013.		Small ( esthetic) changes
// Version 5.4

#include <afxwin.h>

#include <afxsock.h>	// CSocket Class
#include <afxmt.h>		// Mutex
#include <ctype.h>		// isdigit()
#include <stdio.h>
#include <stdlib.h>		// atoi()
#include <string.h>		// strcpy(), strlen()

#include "CEngineDef.h"
#include "CEngineDefinitions.h"

#include "CMoveGen64.h"
#include "CBook64.h"
#include "CDatabase64.h"
#include "CSearch64.h"

#ifdef _ENGINE_DAMAGE
#include "CEvaluation64.h"
#endif

#include "CEngine.h"
#include "pdn.h"
#include "CGuide.h"


inline int imax(int ix, int iy)
{
	return ( ix > iy ? ix : iy ) ;
}


inline int imin(int ix, int iy)
{
	return ( ix < iy ? ix : iy ) ;
}


CEngine::CEngine()
{
	int i ;


	m_pSearch64	= new CSearch64 ;

	// Init a few m_EngineData Parameters
	m_EngineData.bSearch = false ;

	for ( i=0; i<SEARCHTHREAD_MAXTHREAD; i++ )
		m_EngineData.bPPSearch[i] = false ;
	
	m_EngineData.iOutputStackRead	= 0 ;
	m_EngineData.iOutputStackWrite	= 0 ;

	// Init Board
	BoardInit() ;

}


CEngine::~CEngine()
{

	delete m_pSearch64 ;

}


__int64 CEngine::Perft(int iDepth, PTSTACK pTDepth)
{
	// No Bulk Counting implementation
	int i, iXMoves ;
	__int64 iXNodes ;


	if ( iDepth == 0 )
		return ( 1 ) ;

	iXMoves = m_pSearch64->m_pMoveGenerate64->MoveGenerate(pTDepth) ;

	iXNodes				= 0 ;
	(pTDepth+1)->bTurn	= !pTDepth->bTurn ;

	for ( i=0; i<iXMoves; i++ ) {

		m_pSearch64->Search_Move64(pTDepth,pTDepth->FirstMove++ ) ;
		iXNodes += Perft(iDepth-1, pTDepth+1) ;
	}

	return ( iXNodes ) ;
}


void CEngine::BoardLog( int* pBoard, bool bTurn, CString* pcsBoard)
{
	int i, j, iPosition ;


	pcsBoard->Empty() ;
	pcsBoard->Format("\n") ;

	for ( i=0; i<10; i++ ) {
		for ( j=0; j<5; j++ ) {

			iPosition = 5 * i + j + 1 ;

			if ( pBoard[iPosition] ==  ENGINE_WHITEMAN ) {

				if ( i%2 )
					*pcsBoard += " w ." ;
				else
					*pcsBoard += " . w" ;

			} else if ( pBoard[iPosition] ==  ENGINE_WHITEKING ) {

				if ( i%2 )
					*pcsBoard += " W ." ;
				else
					*pcsBoard += " . W" ;

			} else if ( pBoard[iPosition] ==  ENGINE_BLACKMAN ) {
				
				if ( i%2 )
					*pcsBoard += " b ." ;
				else
					*pcsBoard += " . b" ;

			} else if ( pBoard[iPosition] ==  ENGINE_BLACKKING ) {

				if ( i%2 )
					*pcsBoard += " B ." ;
				else
					*pcsBoard += " . B" ;

			} else
				*pcsBoard += " . ." ;
		}

		*pcsBoard += "\n" ;
	}

	if ( bTurn == true )
		*pcsBoard += "White Turn\n" ;
	else
		*pcsBoard += "Black Turn\n" ;
}


bool CEngine::DamageLog(int iMode)
{
	int iNGames ;
	float fTime ;
	CString csBoardLog, csGameResult, csTime ;
	clock_t ctGameTime ;
	time_t ttGameTime ;


	if ( m_EngineData.iSPOutput == 0 )
		return ( false ) ;

	iNGames = m_EngineData.DXPMatch.iNGames ;

	switch ( iMode ) {

		case DXP_MATCH_LOG_OPEN:			// Open Match LogFile

			if ( fopen_s(&m_EngineData.DXPMatch.pMatchLogFile, m_EngineData.DXPMatch.csMatchLogFile, "w") == 0 )
				m_EngineData.DXPMatch.bMatchLogFileOpen = true ;
			else
				m_EngineData.DXPMatch.bMatchLogFileOpen = false ;

			break ;

		case DXP_MATCH_LOG_GAME_START:		// Write Game Start Info to Match LogFile

			ttGameTime	= time(NULL) ;
			ctGameTime	= clock() ;

			fprintf(m_EngineData.DXPMatch.pMatchLogFile,"\nGame: %d, Start: %s", iNGames + 1, ctime(&ttGameTime) ) ;
			
			break ;

		case DXP_MATCH_LOG_GAME_BOARD:		// Write Game Move Board to Match LogFile

			BoardLog(m_EngineData.iBoard, m_EngineData.bTurn, &csBoardLog) ;

			fprintf(m_EngineData.DXPMatch.pMatchLogFile,"%s\n", csBoardLog ) ;
			
			break ;

		case DXP_MATCH_LOG_GAME_MOVE:		// Write Game Move Info to Match LogFile

			fTime = DamageTime(DXP_MATCH_GAME_MOVE) ;

			FormatDamageTime(fTime, &csTime) ;

			fprintf(m_EngineData.DXPMatch.pMatchLogFile,"\nMove: %d ( Game %d ), Time = %s\n",  m_EngineData.iGameHistoryActualMove/2, ( m_EngineData.DXPMatch.iNGames + 1 ), csTime ) ;
	
			break ;

		case DXP_MATCH_LOG_GAME_END:		// Write Game End Info to Match LogFile

			ttGameTime	= time(NULL) ;
			
			fTime = DamageTime(DXP_MATCH_GAME_MOVE) ;

			FormatDamageTime(fTime, &csTime) ;

			DXPGameFormat(&csGameResult) ;

			fprintf(m_EngineData.DXPMatch.pMatchLogFile,"\nGame: %d %s, Time: %s, End: %s", iNGames + 1, csGameResult, csTime, ctime(&ttGameTime) ) ;
	
			break ;

		case DXP_MATCH_LOG_STAT:			// Write DXP Stat Info to Match LogFile

			fprintf(	m_EngineData.DXPMatch.pMatchLogFile,"DXP Statistics: Games = %d, Win = %d, Lost = %d, Draw = %d, Unknown = %d\n", 
						iNGames, m_EngineData.DXPMatch.iNWin, m_EngineData.DXPMatch.iNLoose, m_EngineData.DXPMatch.iNDraw, m_EngineData.DXPMatch.iNUnknown) ;
			break ;
	}

	return ( true ) ;
}


float CEngine::DamageTime(int iMode)
{

	time_t iNSeconds, iNMilliSeconds ;
	float fTime ;
	struct timeb timebTime ;

	ftime(&timebTime) ;

	fTime = NULL ;

	if ( iMode == DXP_MATCH_GAME_START ) {

		m_EngineData.DXPMatch.DXPGameStart = timebTime ;

	}

	if ( iMode == DXP_MATCH_GAME_MOVE ) {

		iNSeconds		= timebTime.time -  m_EngineData.DXPMatch.DXPGameStart.time ;
		iNMilliSeconds	= timebTime.millitm - m_EngineData.DXPMatch.DXPGameStart.millitm ;

		fTime		= (float)( ( 1000 * iNSeconds )  + iNMilliSeconds ) ;
		fTime	   /= 1000. ;

	}

	if ( iMode == DXP_MATCH_GAME_END ) {

		m_EngineData.DXPMatch.DXPGameEnd = timebTime ;

	}

	return ( fTime ) ;
}


void CEngine::FormatDamageTime(float fTime, CString* pcsTime) 
{
	int iHour, iMin, iSec, iTime ;

	iTime	= (int)fTime ;

	iHour	= iTime / 3600 ;
	iMin	= iTime / 60 ;
	iSec	= iTime % 60 ;

	pcsTime->Format("%02dh:%02dm:%02ds", iHour, iMin, iSec) ;
}


void CEngine::ClearOutput(int iOutput)
{
	CString csMessage ;

	csMessage.Empty() ;
	SendEngineMessage(iOutput, ENGINE_$CLEAROUTPUT, csMessage) ;

}


void CEngine::EvalOutput(void)
{

#ifdef _ENGINE_DAMAGE

	int i, iCoreOutput ;
	CString csEvaluation ;
	PTSTACK pTDepth ;

	iCoreOutput = 1 ;

	ClearOutput(iCoreOutput) ;

	pTDepth = m_EngineData.SearchThreadData[0].pTreeStack ;

	for ( i=0 ; i<5 ; i++ )		// 64Bit
		pTDepth->bbField[i] = m_EngineData.bbField[i] ;

	m_pSearch64->MaterialValue(pTDepth) ;

	m_pSearch64->m_pEvaluation64->Eval64_Position_Info(pTDepth->bTurn, pTDepth, &csEvaluation) ;

	SendEngineMessage(iCoreOutput, ENGINE_$EVAL, csEvaluation) ;

#endif

}


bool CEngine::FENRead(char* cBuf)
{
	bool bTurn ;
	char cChar ;
	int i, iBoard[51], iColor, iDigit0, iDigit1, iFen, iLength, iPiece, iPosition ;
	int iNWhiteMan, iNWhiteKing, iNBlackMan, iNBlackKing ;


	for ( i=0; i<51; i++ )
		iBoard[i] = 0 ;

	if ( cBuf[0] == 'W' )
		bTurn = true ;
	else if ( cBuf[0] == 'B' )
		bTurn = false ;
	else
		return ( false ) ;		// FEN should start with W or B

	if ( cBuf[1] != ':' )
		return ( false ) ;		// FEN[1] should contain ':' character

	iLength = (int)strlen(cBuf) ;

	if ( cBuf[iLength-2] != '.' )
		return ( false ) ;		// FEN should end with '.' character

	iFen	= 2 ;
	iColor	= -1 ;				// Non-Existing Color

	iNWhiteMan	= 0 ;
	iNWhiteKing = 0 ;
	iNBlackMan	= 0 ;
	iNBlackKing = 0 ;

	do {

		cChar = cBuf[iFen] ;

		if ( cChar == 'W' ) {

			iColor	= WHITE ;
			iPiece = WHITEMAN ;

		} else if ( cChar == 'B' ) {

			iColor = BLACK ;
			iPiece = BLACKMAN ;

		} else if ( cChar == 'K' ) {

			if ( iColor == WHITE )
				iPiece = WHITEKING ;
			else if ( iColor == BLACK ) 
				iPiece = BLACKKING ;
			else
				return ( false ) ;	// Piece identification can not start without Color Defined 
		
		} else if ( cChar == ',' ) {

			if ( iColor == WHITE )
				iPiece = WHITEMAN ;
			else if ( iColor == BLACK )
				iPiece = BLACKMAN ;
			else
				return ( false ) ;	// Character ',' can not occur before Color Defined 
		
		} else if ( isdigit(cChar) ) {

			iDigit0 = atoi(&cChar) ;

			cChar = cBuf[iFen+1] ;

			if ( isdigit(cChar) ) {		// Position has 2 characters

				iDigit1		= atoi(&cChar) ;
				iPosition	= ( 10 * iDigit0 ) + iDigit1 ;
				
				++iFen ;

			} else {					// Position has 1 character

				iPosition = iDigit0 ;
			}

			iBoard[iPosition] = iPiece ;

			if ( iPiece == WHITEMAN )
				++iNWhiteMan ;
			else if ( iPiece == WHITEKING )
				++iNWhiteKing ;
			else if ( iPiece == BLACKMAN )
				++iNBlackMan ;
			else if ( iPiece == BLACKKING )
				++iNBlackKing ;
		}

		++iFen ;

	} while ( cBuf[iFen] != '.' ) ;

	for ( i=0; i<51; i++ )
		m_EngineData.iBoard[i] = iBoard[i] ;

	m_pSearch64->ConvertBoard(&m_EngineData) ;

	m_EngineData.bTurn = bTurn ;

	return ( true ) ;
}


bool CEngine::PDNCopyBoard(void)
{
	int i ;

	m_EngineData.bTurn = m_pPDN->m_bTurn ;

	for ( i=1; i<=50; i++ )
		m_EngineData.iBoard[i] = m_pPDN->m_iBoard[i] ;

	m_pSearch64->ConvertBoard(&m_EngineData) ;

	return ( true ) ;
}


bool CEngine::PDNCopyMoves(void)
{
	int i, j, iFindMove, iMoveData[2] ;
	CString csMessage ;
	PTSTACK pTDepth ;

	pTDepth = m_EngineData.pTreeStack ;

	pTDepth->bTurn = m_EngineData.bTurn ;

	for ( i=0; i<5; i++ )
		pTDepth->bbField[i] = m_EngineData.bbField[i] ;

	m_pSearch64->MaterialValue(pTDepth) ;

	m_EngineData.iGameHistoryActualMove	= 0 ;
	m_EngineData.iGameHistoryMaxMove	= 0 ;
	
	m_EngineData.GameHistory[0].bTurn = m_EngineData.bTurn ;

	for ( i=0 ; i<5 ; i++ )		// 64Bit
		m_EngineData.GameHistory[0].bbField[i] = m_EngineData.bbField[i] ;	


	for ( j=0; j<m_iPDNMoves; j++ ) {

		iMoveData[0] = m_pPDN->m_iMovePDN[j][1] ; 
		iMoveData[1] = m_pPDN->m_iMovePDN[j][2] ;

		iFindMove = m_pSearch64->Find_Move(iMoveData, m_EngineData.pTreeStack, m_EngineData.pMoveStack) ;

		if ( iFindMove == -1 ) {

			csMessage.Format("Wrong Move %d!", j) ;
			SendEngineMessage(ENGINE_CONSOLE, ENGINE_$PDN, csMessage) ;

			return ( false ) ;
		}

		m_pSearch64->Search_Move64(pTDepth, pTDepth->FirstMove + iFindMove) ;

		for ( i=0; i<4; i++ )
			pTDepth->bbField[i] = (pTDepth+1)->bbField[i] ;

		pTDepth->bTurn = !(pTDepth->bTurn) ;

		// Update GameHistory !
		m_EngineData.GameHistory[0].bTurn = pTDepth->bTurn ;

		for ( i=0 ; i<5 ; i++ )		// 64Bit
			m_EngineData.GameHistory[j+1].bbField[i] = pTDepth->bbField[i] ;	

		m_EngineData.iGameHistoryActualMove	= j ;
		m_EngineData.iGameHistoryMaxMove	= j ;
	}

	for ( i=0; i<5; i++ )
		m_EngineData.bbField[i] = pTDepth->bbField[i] ;

	m_EngineData.bTurn = pTDepth->bTurn ;

	m_pSearch64->DeConvertBoard64(&m_EngineData) ;

	return ( true ) ;
}


void CEngine::MoveToDXPString(CString* csMove)
{
	int i ;
	CString csCapture ;

	csMove->Format("%.2d%.2d%.2d", m_iMoveData[0], m_iMoveData[1], m_iMoveData[2] ) ;

	for ( i = 0; i < m_iMoveData[2] ; i++ ) {
		csCapture.Format("%.2d", m_iMoveData[i + 3] ) ;	// Position Capture Piece
		*csMove += csCapture ;
	}
}


void CEngine::MoveToString(bool bNotation, CString* csMove)
{
	int i ;
	CString csCapture ;

	if ( bNotation ) {

		if (  m_iMoveData[2] > 0 )
			csMove->Format("%dx%d", m_iMoveData[0], m_iMoveData[1] ) ;
		else
			csMove->Format("%d-%d", m_iMoveData[0], m_iMoveData[1] ) ;

	} else {

		csMove->Format("%.2d %.2d %.2d", m_iMoveData[0], m_iMoveData[1], m_iMoveData[2] ) ;

		for ( i = 0; i < m_iMoveData[2] ; i++ ) {
			csCapture.Format(" %.2d", m_iMoveData[i + 3] ) ;	// Position Capture Piece
			*csMove += csCapture ;
		}
	}
}


void CEngine::MoveListToString(CString* pcsMoveList)
{
	int i, iXMoves, iMoveData[23], j ;
	CString csMove, csCapture ;
	PTSTACK pTDepth ;

	pTDepth = m_EngineData.pTreeStack ;

	pTDepth->bTurn = m_EngineData.bTurn ;

	iXMoves = m_pSearch64->m_pMoveGenerate64->MoveGenerate(pTDepth ) ;

	for ( i=0; i<iXMoves; i++ ) {
		m_pSearch64->Decode_Move(i, iMoveData, m_EngineData.pTreeStack, m_EngineData.pMoveStack) ;

		csMove.Format(" m %.2d %.2d %.2d", iMoveData[0], iMoveData[1], iMoveData[2] ) ;

		for ( j = 0; j < iMoveData[2] ; j++ ) {
			csCapture.Format(" %.2d", iMoveData[j +3]) ;
			csMove += csCapture ;
		}

		*pcsMoveList += csMove ;
	}
}


bool CEngine::DXPBallot(int iGameBallot, CString* pcsDXPPosition, CString* pcsWhiteMove, CString* pcsBlackMove)
{
	int i, iIndex, iWhiteFrom, iWhiteTo, iBlackFrom, iBlackTo ;

	if ( iGameBallot < 0 || iGameBallot > 157 )
		return ( false ) ;

	m_bDXPClientTurn = ( iGameBallot%2 ? true : false ) ;

	m_EngineData.DXPMatch.bDXPClientTurn = m_bDXPClientTurn ;

	iGameBallot /=2 ;									// Every Game for 2 colors to move !

	// Init Board
	for ( i=1; i<=20; i++ )
		m_EngineData.iBoard[i] = ENGINE_BLACKMAN ;		// Black Man

	for ( i=21; i<=30; i++ )
		m_EngineData.iBoard[i] = 0 ;					// Empty

	for ( i=31; i<=50; i++ )
		m_EngineData.iBoard[i] = ENGINE_WHITEMAN ;		// White Man

	iIndex = 4 * iGameBallot ;

	iWhiteFrom	= iBallot[iIndex] ;
	iWhiteTo	= iBallot[iIndex+1] ;
	iBlackFrom	= iBallot[iIndex+2] ;
	iBlackTo	= iBallot[iIndex+3] ;

	pcsWhiteMove->Format("%d-%d", iWhiteFrom, iWhiteTo) ;
	pcsBlackMove->Format("%d-%d", iBlackFrom, iBlackTo) ;

	m_EngineData.iBoard[iWhiteFrom]	= 0 ;
	m_EngineData.iBoard[iWhiteTo]	= ENGINE_WHITEMAN ;
	m_EngineData.iBoard[iBlackFrom]	= 0 ;
	m_EngineData.iBoard[iBlackTo]	= ENGINE_BLACKMAN ;

	return (true ) ;
}


bool CEngine::DXPBallotGame(int iBallot)
{
	int i, iWhiteMove, iBlackMove ;
	PTSTACK pTDepth ;
	CString csPosition, csWhiteMove, csBlackMove ;

	m_EngineData.bTurn = true ;		// Ballot Game starts with White to Move

	if ( m_EngineData.DXPMatch.iMatchMode == 0 ) {

		DXPBallot(iBallot,&csPosition, &csWhiteMove, &csBlackMove) ;

		m_EngineData.DXPMatch.csBallotWhiteMove = csWhiteMove ;
		m_EngineData.DXPMatch.csBallotBlackMove = csBlackMove ;

		m_pSearch64->ConvertBoard(&m_EngineData) ;

	} else {

		m_bDXPClientTurn = ( iBallot%2 ? true : false ) ;
		m_EngineData.DXPMatch.bDXPClientTurn = m_bDXPClientTurn ;

		pTDepth	= m_EngineData.pTreeStack ;

		for ( i=0; i<5; i++ ) {
			m_EngineData.bbField[i] = m_EngineData.DXPMatch.bbField[i] ;
			pTDepth->bbField[i]		= m_EngineData.DXPMatch.bbField[i] ;
		}

		iWhiteMove = ( iBallot / 2 ) / m_EngineData.DXPMatch.iWhiteMoves ;
		iBlackMove = ( iBallot / 2 ) % m_EngineData.DXPMatch.iWhiteMoves ;

		pTDepth->bTurn	= true ;
		m_pSearch64->m_pMoveGenerate64->MoveGenerate(pTDepth) ;
		m_pSearch64->Format_Move64(pTDepth->FirstMove, iWhiteMove, &csWhiteMove) ;
		m_pSearch64->Search_Move64(pTDepth, pTDepth->FirstMove + iWhiteMove) ;

		for ( i=0; i<5; i++ )
			pTDepth->bbField[i] = (pTDepth+1)->bbField[i] ;

		pTDepth->bTurn	= false ;
		m_pSearch64->m_pMoveGenerate64->MoveGenerate(pTDepth) ;
		m_pSearch64->Format_Move64(pTDepth->FirstMove, iBlackMove, &csBlackMove) ;
		m_pSearch64->Search_Move64(pTDepth, pTDepth->FirstMove + iBlackMove) ;

		for ( i=0; i<5; i++ )
			pTDepth->bbField[i] = (pTDepth+1)->bbField[i] ;

		for ( i=0; i<5; i++ )
			m_EngineData.bbField[i] = pTDepth->bbField[i] ;

		m_pSearch64->DeConvertBoard64(&m_EngineData) ;

	}

	// Reset Game History
	m_EngineData.iGameHistoryActualMove	= 0 ;
	m_EngineData.iGameHistoryMaxMove	= 0 ;
	
	m_EngineData.GameHistory[0].bTurn = m_EngineData.bTurn ;

	for ( i=0 ; i<5 ; i++ )		// 64Bit
		m_EngineData.GameHistory[0].bbField[i] = m_EngineData.bbField[i] ;	

	// Temporary implementation / Hack
	m_EngineData.GameHistory[0].csMove = csWhiteMove ;
	m_EngineData.GameHistory[1].csMove = csBlackMove ;

	m_EngineData.iGameHistoryActualMove	= 2 ;	// Ballot Game, 2 moves already defined !!

	return ( true ) ;
}


bool CEngine::DXPUpdateStat(int iGame)
{
	int iGameResult ;


	iGameResult = m_EngineData.DXPMatch.iGameResult ;
		
	m_EngineData.DXPMatch.iResult[iGame]		= iGameResult ;
	m_EngineData.DXPMatch.iClientResult[iGame]	= iGameResult ; 

	if ( iGameResult == DXP_WIN )
		++ m_EngineData.DXPMatch.iNWin ;
	else if ( iGameResult == DXP_DRAW )
		++ m_EngineData.DXPMatch.iNDraw ;
	else if ( iGameResult == DXP_LOSE )
		++ m_EngineData.DXPMatch.iNLoose ;
	else if ( iGameResult == DXP_UNKNOWN )
		++ m_EngineData.DXPMatch.iNUnknown ;

	return ( true ) ;
}


void CEngine::DXPGameFormat(CString* pcsGameResult)
{
	int iGameResult ;


	iGameResult = m_EngineData.DXPMatch.iGameResult ;
		
	if ( iGameResult == DXP_WIN )
		pcsGameResult->Format("Win") ;
	else if ( iGameResult == DXP_DRAW )
		pcsGameResult->Format("Draw") ;
	else if ( iGameResult == DXP_LOSE )
		pcsGameResult->Format("Lost") ;
	else if ( iGameResult == DXP_UNKNOWN )
		pcsGameResult->Format("Unknown") ;
}


void CEngine::DXPStat(void)
{
	CString csMessage ;

	csMessage.Format("Games   = %d", m_EngineData.DXPMatch.iNGames) ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXPSTAT_MESSAGE, csMessage) ;

	csMessage.Format("Win     = %d", m_EngineData.DXPMatch.iNWin) ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXPSTAT_MESSAGE, csMessage) ;

	csMessage.Format("Loose   = %d", m_EngineData.DXPMatch.iNLoose) ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXPSTAT_MESSAGE, csMessage) ;

	csMessage.Format("Draw    = %d", m_EngineData.DXPMatch.iNDraw) ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXPSTAT_MESSAGE, csMessage) ;

	csMessage.Format("Unknown = %d", m_EngineData.DXPMatch.iNUnknown) ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXPSTAT_MESSAGE, csMessage) ;


}


bool CEngine::DXPStatMatchFile(CString* pcsDXPMatchFile)
{
	bool bClientWhite ;
	int iNGames, iNLines ;
	int iNWin, iNDraw, iNLose, iNUnknown ;
	CString csMessage, csEngineName, csTAGInfo ;
	CFile DXPMatchFile ;

	if ( !DXPMatchFile.Open(*pcsDXPMatchFile, CFile::modeRead) )
		return ( false ) ;

	CArchive CArchive_DXPFile(&DXPMatchFile, CArchive::load) ;

	iNGames		= 0 ;
	iNLines		= 0 ;

	iNWin		= 0 ;
	iNDraw		= 0 ;
	iNLose		= 0 ;
	iNUnknown	= 0 ;

	csEngineName = m_EngineData.csEngineName ;

	while ( CArchive_DXPFile.ReadString(csTAGInfo) ) {
	
		if ( csTAGInfo.Find("Event") != -1 )
			++iNGames ;

		if ( csTAGInfo.Find("White") != -1 )
			if ( csTAGInfo.Find(csEngineName) != -1 )
				bClientWhite = true ;

		if ( csTAGInfo.Find("Black") != -1 )
			if ( csTAGInfo.Find(csEngineName) != -1 )
				bClientWhite = false ;

		if ( csTAGInfo.Find("Result") != -1 ) {

			if ( csTAGInfo.Find("2-0") != -1 )
				if ( bClientWhite )
					++iNWin ;
				else
					++iNLose ;

			if ( csTAGInfo.Find("1-1") != -1 )
				++iNDraw ;

			if ( csTAGInfo.Find("0-2") != -1 )
				if ( bClientWhite )
					++iNLose ;
				else
					++iNWin ;
				
			if ( csTAGInfo.Find("?") != -1 )
				++iNUnknown ;
		}

		++iNLines ;
	}


	CArchive_DXPFile.Close() ;
	DXPMatchFile.Close() ;

	csMessage.Format("Games   = %d", iNGames) ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXPSTAT_MESSAGE, csMessage) ;

	csMessage.Format("Win     = %d", iNWin) ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXPSTAT_MESSAGE, csMessage) ;

	csMessage.Format("Loose   = %d", iNLose) ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXPSTAT_MESSAGE, csMessage) ;

	csMessage.Format("Draw    = %d", iNDraw) ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXPSTAT_MESSAGE, csMessage) ;

	csMessage.Format("Unknown = %d", iNUnknown) ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXPSTAT_MESSAGE, csMessage) ;

	return ( true ) ;
}


bool CEngine::DXPPositionToString(CString* pcsDXPPosition)
{
	int i,iType ;

	pcsDXPPosition->Empty() ;

	for ( i=1; i<=50; i++ ) {

		iType = m_EngineData.iBoard[i] ;

		if ( iType == ENGINE_WHITEMAN )
			*pcsDXPPosition += 'w' ;

		else if ( iType == ENGINE_WHITEKING )
			*pcsDXPPosition += 'W' ;

		else if ( iType == ENGINE_BLACKMAN )
			*pcsDXPPosition += 'z' ;

		else if ( iType == ENGINE_BLACKKING )
			*pcsDXPPosition += 'Z' ;

		else if ( iType == ENGINE_EMPTY )
			*pcsDXPPosition += 'e' ;
	}

	return true ;
}


bool CEngine::DXPStringToPosition(bool bTurn, CString* pcsDXPPosition)
{

	int i, iType ;
	CString csToken ;

	m_EngineData.bTurn = bTurn;

	for ( i=0; i<50; i++ ) {

		csToken = pcsDXPPosition->Mid(44+i, 1) ;

		if ( csToken == 'w' )
			iType = ENGINE_WHITEMAN ;

		else if ( csToken == 'W' )
			iType = ENGINE_WHITEKING ;

		else if ( csToken == 'z' )
			iType = ENGINE_BLACKMAN ;

		else if ( csToken == 'Z' )
			iType = ENGINE_BLACKKING ;	
		
		else if ( csToken == 'e' )
			iType = ENGINE_EMPTY ;
		
		m_EngineData.iBoard[i+1] = iType ;
	}

	return true ;
}


void CEngine::PositionToString(CString* csPosition)
{
	int i ;

	csPosition->Empty() ;

	for ( i=1 ; i<=50 ; i++ ) {
		if ( m_EngineData.iBoard[i] == ENGINE_WHITEMAN )
			*csPosition += 'w' ;
		else if ( m_EngineData.iBoard[i] == ENGINE_WHITEKING )
			*csPosition += 'W' ;
		else if ( m_EngineData.iBoard[i] == ENGINE_BLACKMAN )
			*csPosition += 'b' ;
		else if ( m_EngineData.iBoard[i] == ENGINE_BLACKKING )
			*csPosition += 'B' ;
		else
			*csPosition += 'e' ;
	}
}


bool CEngine::StringToPosition(bool bTurn, CString* pcsPosition)
{
	int i, iPos, iType ;
	CString csToken ;

	m_EngineData.bTurn = bTurn;

	if ( pcsPosition->GetLength() != 50 ) {

		while ( ParseMessage(pcsPosition, &csToken) ) {

			if ( isdigit(csToken[0]) ) {
				iPos = atoi( LPCTSTR(csToken) ) ;	
				m_EngineData.iBoard[iPos] = iType ;
		
			} else {

				if ( csToken == ENGINE_STR_WHITEMAN )
					iType = ENGINE_WHITEMAN ;

				else if ( csToken == ENGINE_STR_WHITEKING )
					iType = ENGINE_WHITEKING ;

				else if ( csToken == ENGINE_STR_BLACKMAN )
					iType = ENGINE_BLACKMAN ;

				else if ( csToken == ENGINE_STR_BLACKKING )
					iType = ENGINE_BLACKKING ;

				else if ( csToken == ENGINE_STR_EMPTY )
					iType = ENGINE_EMPTY ;
				}			

			}

		return ( true ) ;
	}

	for ( i=0 ; i<50 ; i++ ) {

		if ( pcsPosition->GetAt(i) == ENGINE_STR_CHAR_WHITEMAN )
			iType = ENGINE_WHITEMAN ;

		else if ( pcsPosition->GetAt(i) == ENGINE_STR_CHAR_WHITEKING )
			iType = ENGINE_WHITEKING ;

		else if ( pcsPosition->GetAt(i) == ENGINE_STR_CHAR_BLACKMAN )
			iType = ENGINE_BLACKMAN ;

		else if ( pcsPosition->GetAt(i) == ENGINE_STR_CHAR_BLACKKING )
			iType = ENGINE_BLACKKING ;
		
		else if ( pcsPosition->GetAt(i) == ENGINE_STR_CHAR_EMPTY )
			iType = ENGINE_EMPTY ;
		else
			return ( false ) ;
		
		m_EngineData.iBoard[i+1] = iType ;

	}

	return ( true ) ;
}


void CEngine::DXPClientStart(void)
{
	bool bInitBoard ;
	int i, iLength, iNMoves, iTime ;
	CString csDXPPosition, csEngineName, csMessage, csMoves, csTime ;

	iNMoves = m_EngineData.DXPMatch.iMaxMoves ;
	iTime	= m_EngineData.DXPMatch.iMaxTime ;

	bInitBoard = true ;

	for ( i=1; i<=50; i++ )
		if ( m_EngineData.iBoard[i] != iInitBoard[i] ) {

			bInitBoard= false ;
			break ;
		}

	ClearOutput(0) ;

	csEngineName = m_EngineData.csEngineName ;

	iLength = csEngineName.GetLength() ;

	for ( i=0; i < ( DXP_NAMELENGTH - iLength ) ; i++ )
		csEngineName.AppendChar(' ') ;

	iLength = csEngineName.GetLength() ;

	csMoves.Format("%03d", m_EngineData.DXPMatch.iMaxMoves ) ;	// Max Moves
	csTime.Format("%03d", m_EngineData.DXPMatch.iMaxTime ) ;	// Max Time

	csMessage.Format("%s", ENGINE_DXP_STR_GAMEREQ) ;

	csMessage += DXP_VERSION ; // Add DXP Version
	csMessage += csEngineName ;

	if ( m_bDXPClientTurn )
		csMessage += DXP_BLACK ;
	else
		csMessage += DXP_WHITE ;

	csMessage += csTime ;
	csMessage += csMoves ;

	if ( bInitBoard )
		csMessage += DXP_INITPOSITION ;
	
	else {

		csMessage += DXP_DXPPOSITION ;
		
		if ( m_EngineData.bTurn )
			csMessage += DXP_WHITE ;
		else
			csMessage += DXP_BLACK ;

		DXPPositionToString(&csDXPPosition) ;

		csMessage+= csDXPPosition ;
	}

	iLength = csMessage.GetLength() ;


	m_EngineMessage[m_iNMessage].bDisplayUpdate	= true ;
	
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXPGAMEREQ_MESSAGE, csMessage) ;
}


int CEngine::DecodeDXPMessage(CString csMessage)
{
	int iMessageID ;

	iMessageID = -1 ;

	if ( csMessage.GetAt(0) == 'R')				// GameReQ
		iMessageID = 1 ;

	else if ( csMessage.GetAt(0) == 'A' )		// GameAcc
		iMessageID = 2 ;

	else if ( csMessage.GetAt(0) == 'M' )		// Move
		iMessageID = 3 ;

	else if ( csMessage.GetAt(0) == 'E' )		// GameEnd
		iMessageID = 4 ;

	else if ( csMessage.GetAt(0) == 'C' )		// Chat
		iMessageID = 5 ;

	else if ( csMessage.GetAt(0) == 'B' )		// BackReq
		iMessageID = 6 ;

	else if ( csMessage.GetAt(0) == 'K' )		// BackAcc
		iMessageID = 7 ;

	m_EngineMessage[m_iNMessage].iMessageID		= iMessageID ;
	m_EngineMessage[m_iNMessage].bDisplayUpdate	= false ;
	m_EngineMessage[m_iNMessage].csMessageBody	= csMessage ;

	return ( iMessageID ) ;
}


int CEngine::DecodeMessage(CString csMessage)
{
	CString csDebugMessage, csMessageCommand, csMessageBody ;
	int iMessageID ;

	if ( m_iNMessage >= ENGINE_MAX_MESSAGE )
		return ENGINE_MESSAGE_BUFFER_FULL ;	// Message Buffer is full, in fact a warning should be generated !

	if ( m_bDebug ) {	// Reflect Message !

		csDebugMessage.Format("%s %s", ENGINE_STR_$DEBUG, csMessage ) ;
		SendEngineMessage(ENGINE_CONSOLE, ENGINE_$DEBUG, csDebugMessage) ;

	}

	if ( !CommandMessage(csMessage, &csMessageCommand, &csMessageBody) )
		return ENGINE_ILLEGAL_COMMAND ;

	iMessageID = -1 ;

	if ( csMessageCommand == ENGINE_STR_ENGINEMODE )
		iMessageID = ENGINE_ENGINEMODE ;

	else if ( csMessageCommand == ENGINE_STR_FORWARD )
		iMessageID = ENGINE_FORWARD ;

	else if ( csMessageCommand == ENGINE_STR_GAME )
		iMessageID = ENGINE_GAME ;

	else if ( csMessageCommand == ENGINE_STR_GO )
		iMessageID = ENGINE_GO ;

	else if ( csMessageCommand == ENGINE_STR_HALT )
		iMessageID = ENGINE_HALT ;

	else if ( csMessageCommand == ENGINE_STR_JUMPTO )
		iMessageID = ENGINE_JUMPTO ;

	else if ( csMessageCommand == ENGINE_STR_LEVELS_BLITZGAME )
		iMessageID = ENGINE_LEVELS_BLITZGAME ;

	else if ( csMessageCommand == ENGINE_STR_LEVELS_LONGGAME )
		iMessageID = ENGINE_LEVELS_LONGGAME ;

	else if ( csMessageCommand == ENGINE_STR_LEVELS_FIXEDDEPTH )
		iMessageID = ENGINE_LEVELS_FIXEDDEPTH ;

	else if ( csMessageCommand == ENGINE_STR_LEVELS_FIXEDTIME )
		iMessageID = ENGINE_LEVELS_FIXEDTIME ;

	else if ( csMessageCommand == ENGINE_STR_LEVELS_INFINITEANALYSIS )
		iMessageID = ENGINE_LEVELS_INFINITEANALYSIS ;

	else if ( csMessageCommand == ENGINE_STR_MOVE )
		iMessageID = ENGINE_MOVE ;

	else if ( csMessageCommand == ENGINE_STR_NEW )
		iMessageID = ENGINE_NEW ;

	else if ( csMessageCommand == ENGINE_STR_OFFERDRAW )
		iMessageID = ENGINE_OFFERDRAW ;

	else if ( csMessageCommand == ENGINE_STR_OPPONENTS )
		iMessageID = ENGINE_OPPONENTS ;

	else if ( csMessageCommand == ENGINE_STR_PING )
		iMessageID = ENGINE_PING ;

	else if ( csMessageCommand == ENGINE_STR_POSITION )
		iMessageID = ENGINE_POSITION ;

	else if ( csMessageCommand == ENGINE_STR_REQUESTGAME )
		iMessageID = ENGINE_REQUESTGAME ;

	else if ( csMessageCommand == ENGINE_STR_REQUESTPOSITION )
		iMessageID = ENGINE_REQUESTPOSITION ;

	else if ( csMessageCommand == ENGINE_STR_SETCOMMENT )
		iMessageID = ENGINE_SETCOMMENT ;

	else if ( csMessageCommand == ENGINE_STR_SETOPTION )
		iMessageID = ENGINE_SETOPTION ;

	else if ( csMessageCommand == ENGINE_STR_SHOWWINDOW )
		iMessageID = ENGINE_SHOWWINDOW ;

	else if ( csMessageCommand == ENGINE_STR_STATUS )
		iMessageID = ENGINE_STATUS ;

	else if ( csMessageCommand == ENGINE_STR_QUIT )
		iMessageID = ENGINE_QUIT ;

	// S - Commands
	else if ( csMessageCommand == ENGINE_STR_$CLEARCOUNT )
		iMessageID = ENGINE_$CLEARCOUNT ;

	else if ( csMessageCommand == ENGINE_STR_$CLEAROUTPUT )
		iMessageID = ENGINE_$CLEAROUTPUT ;

	else if ( csMessageCommand == ENGINE_STR_$DXP )
		iMessageID = ENGINE_$DXP ;

	else if ( csMessageCommand == ENGINE_STR_$EVAL )
		iMessageID = ENGINE_$EVAL ;	

	else if ( csMessageCommand == ENGINE_STR_$INFO )
		iMessageID = ENGINE_$INFO ;

	else if ( csMessageCommand == ENGINE_STR_$GAME )
		iMessageID = ENGINE_$GAME ;

	else if ( csMessageCommand == ENGINE_STR_$LOG )
		iMessageID = ENGINE_$LOG ;

	else if ( csMessageCommand == ENGINE_STR_$MATCH )
		iMessageID = ENGINE_$MATCH ;

	else if ( csMessageCommand == ENGINE_STR_$PDN )
		iMessageID = ENGINE_$PDN ;

	else if ( csMessageCommand == ENGINE_STR_$PRM )
		iMessageID = ENGINE_$PRM ;

	// $PP -Commands
	else if ( csMessageCommand == ENGINE_STR_$PPMODE )
		iMessageID = ENGINE_$PPMODE ;

	else if ( csMessageCommand == ENGINE_STR_$PPPROC )
		iMessageID = ENGINE_$PPPROC ;

	// !-Commands
	else if ( csMessageCommand == ENGINE_STR_ASX_DB64 )
		iMessageID = ENGINE_ASX_DB64 ;
	
	// ?-Commands
	else if ( csMessageCommand == ENGINE_STR_QMRK_DISPLAYBOARD )
		iMessageID = ENGINE_QMRK_DISPLAYBOARD ;

	// #-Commands
	else if ( csMessageCommand == ENGINE_STR_NMBRSGN_MOVEGENERATE )
		iMessageID = ENGINE_NMBRSGN_MOVEGENERATE ;

	else if ( csMessageCommand == ENGINE_STR_NMBRSGN_PERFT )
		iMessageID = ENGINE_NMBRSGN_PERFT ;

	else if ( csMessageCommand == ENGINE_STR_NMBRSGN_TIME )
		iMessageID = ENGINE_NMBRSGN_TIME ;

	// @-Commands
	else if ( csMessageCommand == ENGINE_STR_AT_HALT )
		iMessageID = ENGINE_AT_HALT ;
	
	else if ( csMessageCommand == ENGINE_STR_AT_OPEN )
		iMessageID = ENGINE_AT_OPEN ;

	else if ( csMessageCommand == ENGINE_STR_AT_LIST )
		iMessageID = ENGINE_AT_LIST ;
	
	else if ( csMessageCommand == ENGINE_STR_AT_PAUSE )
		iMessageID = ENGINE_AT_PAUSE ;
	
	else if ( csMessageCommand == ENGINE_STR_AT_RUN )
		iMessageID = ENGINE_AT_RUN ;

	else if ( csMessageCommand == ENGINE_STR_AT_STOP )
		iMessageID = ENGINE_AT_STOP ;

	m_EngineMessage[m_iNMessage].iMessageID		= iMessageID ;

	if ( iMessageID == ENGINE_QMRK_DISPLAYBOARD  )
		m_EngineMessage[m_iNMessage].bDisplayUpdate	= true ;
	else
		m_EngineMessage[m_iNMessage].bDisplayUpdate	= false ;

	m_EngineMessage[m_iNMessage].csMessageBody  = csMessageBody ;

	if ( iMessageID == -1 )	{	// Engine Message not found.
								// Test specific (non GUI) commands

		if ( csMessageCommand == ENGINE_STR_$DEBUG )
			DLR_Debug() ;

		return ENGINE_USER_COMMAND	;			
	}

	return iMessageID ;
}


bool CEngine::CommandMessage(CString csMessage, CString* pcsMessageCommand, CString* pcsMessageBody )

//	Return true if MessageCommand is found otherwise false

{
	int iFind, iLength ;

	iLength = csMessage.GetLength() ;
	
	if ( iLength == 0 ) {
		pcsMessageCommand->Empty() ;
		pcsMessageBody->Empty() ;

		return false ;
	}

	csMessage.TrimLeft() ;			// Trim spaces

	iFind = csMessage.Find(' ') ;	// Find first space (End of Message Command )

	if ( iFind == -1 )				// No additional command data
		*pcsMessageCommand = csMessage ;
	else
		*pcsMessageCommand = csMessage.Left(iFind) ;

	if ( iFind != -1 ) {
		*pcsMessageBody = csMessage.Right(iLength - iFind) ;
		pcsMessageBody->TrimLeft() ;
	} else
		pcsMessageBody->Empty() ;

	return true ;
}


bool CEngine::ParseMessage(CString* pcsMessage, CString* pcsToken )

//	Return true if Token found, otherwise false

{
	int iFind, iLength ;

	pcsMessage->TrimLeft() ;		// Trim spaces
	pcsToken->Empty() ;				// Empty Token

	iLength = pcsMessage->GetLength() ;

	if ( iLength == 0 )				// No Next Token !
		return false ;

	iFind = pcsMessage->Find(' ') ;	// Find first space (End of Message Command )

	if ( iFind == -1 ) {			// No Next Token !
		*pcsToken = *pcsMessage ;
		pcsMessage->Empty() ;
	} else
		*pcsToken = pcsMessage->Left(iFind) ;

	if ( iFind ) {
		*pcsMessage = pcsMessage->Right(iLength - iFind) ;
		pcsMessage->TrimLeft() ;
	} else
		pcsMessage->Empty() ;

	return true ;
}


bool CEngine::ParseGameMove(CString* pcsMessage, int* iGameMoveData)
{
	int i, iFind, iLength ;
	CString csToken ;

	pcsMessage->TrimLeft() ;			// Trim spaces

	iLength = pcsMessage->GetLength() ;

	if ( iLength == 0 )					// Empty Message !
		return ( false ) ;

	if ( pcsMessage->GetAt(0) != ENGINE_STR_CHAR_GAMEMOVE )
		return ( false ) ;					// GameMove should start with a ENGINE_STR_CHAR_GAMEMOVE

	iFind = pcsMessage->FindOneOf( "0123456789" ) ;	// Find digit	

	if ( iFind == -1 )					// No digit found !
		return ( false ) ;	

	*pcsMessage = pcsMessage->Right(iLength-iFind) ;

	if ( ParseMessage(pcsMessage, &csToken) )	// Move string must contain From Position
		iGameMoveData[0] = atoi( LPCTSTR(csToken) ) ;
	else
		return ( false ) ;

	if ( ParseMessage(pcsMessage, &csToken) )	// Move string must contain To Position
		iGameMoveData[1] = atoi( LPCTSTR(csToken) ) ;
	else
		return ( false ) ;

	if ( ParseMessage(pcsMessage, &csToken) )	
		iGameMoveData[2] = atoi( LPCTSTR(csToken) ) ;
	else								// If Move String des not contain NKill value
		iGameMoveData[2] = -1 ;			// insert -1 as value !

	if ( iGameMoveData[2] > 0 ) 
		for ( i=0 ; i < m_iMoveData[2] ; i++ ) {
			ParseMessage(pcsMessage, &csToken) ;
			iGameMoveData[i + 3] = atoi( LPCTSTR(csToken) ) ;
		}	

	return ( true ) ;
}


void CEngine::BoardClear()
{
	int i ;

	m_EngineData.bTurn = true ;		// White turn

	for ( i=1 ; i<=50 ; i++ )
		m_EngineData.iBoard[i] = EMPTY ;

	m_pSearch64->ConvertBoard(&m_EngineData) ;
}


void CEngine::BoardInit()
{
	int i ;

	m_EngineData.bTurn = true ;		// White turn

	for ( i=1 ; i<=20 ; i++ )
		m_EngineData.iBoard[i] = BLACKMAN ;

	for ( i=21; i<=30 ; i++ )
		m_EngineData.iBoard[i] = EMPTY ;

	for ( i=31 ; i<=50; i++ )
		m_EngineData.iBoard[i] = WHITEMAN ;

	m_pSearch64->ConvertBoard(&m_EngineData) ;

}


bool CEngine::BoardUpdate()
{

	return ( m_EngineMessage[m_iNMessage].bDisplayUpdate ) ;
}


void CEngine::ExecuteDXPMessage(void)
{
	bool bTurn, bEngineTurn ;
	int iDXPMessageID, iNMoves ;
	CString csDXPMessage, csMessage, csServerName ; ;

	iDXPMessageID	= m_EngineMessage[m_iNMessage].iMessageID ;
	csDXPMessage	= m_EngineMessage[m_iNMessage].csMessageBody ;

	switch ( iDXPMessageID ) {

		case ENGINE_DXP_GAMEREQ:	// GameReq

			New() ;
			ClearOutput(0) ;

			csMessage.Format("%s", ENGINE_DXP_STR_GAMEACC) ;

			if ( csDXPMessage.Mid(35,1) == 'W' )
				bEngineTurn = true ;
			else
				bEngineTurn = false ;

			if ( csDXPMessage.Mid(42,1) == 'A' )
				bTurn = true ;

			if ( csDXPMessage.Mid(42,1) == 'B' ) {
			
				if ( csDXPMessage.Mid(43,1) == 'W' )
					bTurn = true ;
				else
					bTurn = false ;

				DXPStringToPosition(bTurn, &csDXPMessage) ;

				m_pSearch64->ConvertBoard(&m_EngineData) ;
			}

			m_bDXPGameEnd = false ;

			SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXP_MESSAGE, csMessage) ;

			if ( bTurn == bEngineTurn )
				Go() ;

			break ;

		case ENGINE_DXP_GAMEACC:	// GameAcc

			csServerName = csDXPMessage.Mid(1,32) ;
			csServerName.TrimRight() ;

			m_EngineData.DXPMatch.csServerName = csServerName ;

			m_bDXPGameEnd = false ;

			m_pGuide->OnDXPGameAcc() ;

			if ( m_EngineData.bTurn == m_bDXPClientTurn )
				Go() ;

			break ;

		case ENGINE_DXP_MOVE:		// Move

			DXPMove() ;

			m_EngineData.pTreeStack->bTurn = m_EngineData.bTurn ;

			if ( m_EngineData.iGameHistoryActualMove >= 2 * m_EngineData.DXPMatch.iMaxMoves ) {

				m_bDXPGameEnd = true ;
				
				csMessage.Format("%s", ENGINE_DXP_STR_GAMEEND) ;
				SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXP_MESSAGE, csMessage) ;
			
			} else if ( ( iNMoves = m_pSearch64->m_pMoveGenerate64->MoveGenerate(m_EngineData.pTreeStack) ) == 0 ) {	// No Moves possible send GAMEEND Command

				m_bDXPGameEnd = true ;

				csMessage.Format("%s", ENGINE_DXP_STR_GAMEEND) ;
				SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXP_MESSAGE, csMessage) ;

			} else {
				
				Go() ;
			}

			break ;

		case ENGINE_DXP_GAMEEND:	// GameEnd

			if ( m_bDXPGameEnd == false ) {

				if ( csDXPMessage == ENGINE_DXP_STR_GAMEEND )
					m_EngineData.DXPMatch.iGameResult = DXP_UNKNOWN ;
				else if ( csDXPMessage == ENGINE_DXP_STR_GAMEEND_LOOSE )
					m_EngineData.DXPMatch.iGameResult = DXP_WIN ;
				else if ( csDXPMessage == ENGINE_DXP_STR_GAMEEND_DRAW )
					m_EngineData.DXPMatch.iGameResult = DXP_DRAW ;
				else if ( csDXPMessage == ENGINE_DXP_STR_GAMEEND_WIN )
					m_EngineData.DXPMatch.iGameResult = DXP_LOSE ;

				m_bDXPGameEnd = true ;

				csMessage.Format("%s", ENGINE_DXP_STR_GAMEEND) ;
				SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXP_MESSAGE, csMessage) ;
			}

			m_pGuide->OnDXPGameEnd() ;

			if ( m_bClient == true ) {

				SetEvent(m_hEventGameStop) ;
			}

			break ;

		case ENGINE_DXP_CHAT:		// Chat
			break ;

		case ENGINE_DXP_BACKREQ:	// BackReq
			break ;

		case ENGINE_DXP_BACKACC:	// BackAcc
			break ;
	}

}


void CEngine::ExecuteMessage()
{
	CString csMessage ;

	switch ( m_EngineMessage[m_iNMessage].iMessageID ) {

		case ENGINE_ENGINEMODE:
			EngineMode() ;
			break ;

		case ENGINE_OPPONENTS:
			Opponents() ;
			break ;

		case ENGINE_LEVELS_LONGGAME:
			LevelsLongGame() ;

		case ENGINE_LEVELS_BLITZGAME:
			LevelsBlitzGame() ;
			break ;

		case ENGINE_LEVELS_FIXEDDEPTH:
			LevelsFixedDepth() ;
			break ;

		case ENGINE_LEVELS_FIXEDTIME:
			LevelsFixedTime() ;
			break ;

		case ENGINE_LEVELS_INFINITEANALYSIS:
			LevelsInfiniteAnalysis() ;
			break ;

		case ENGINE_FORWARD:
			Forward() ;
			break ;

		case ENGINE_GAME:
			Game() ;
			break ;

		case ENGINE_GO:
			Go() ;
			break ;

		case ENGINE_HALT:
			Halt() ;
			break ;

		case ENGINE_JUMPTO:
			JumpTo() ;
			break ;

		case ENGINE_MOVE:
			Move() ;
			break ;

		case ENGINE_NEW:
			New() ;
			break ;

		case ENGINE_OFFERDRAW:
			OfferDraw() ;
			break ;

		case ENGINE_PING:
			Ping() ;
			break ;

		case ENGINE_POSITION:
			Position() ;
			break ;

		case ENGINE_QUIT:
//			Quit Message handling in DraughsEngineApp()
			break ;

		case ENGINE_REQUESTGAME:
			RequestGame() ;
			break ;

		case ENGINE_REQUESTPOSITION:
			RequestPosition() ;
			break ;

		case ENGINE_SETCOMMENT:
			SetComment() ;
			break ;

		case ENGINE_SETOPTION:
			SetOption() ;
			break ;

		case ENGINE_SHOWWINDOW:
			ShowWindow() ;
			break ;

		// S - Commands 
		case ENGINE_$CLEARCOUNT:
			DLR_ClearCount() ;
			break ;

		case ENGINE_$CLEAROUTPUT:
			DLR_ClearOutput() ;
			break ;

		case ENGINE_$DXP:
			DLR_DXP() ;
			break ;

		case ENGINE_$EVAL:
#ifdef _ENGINE_DAMAGE
			DLR_Eval() ;
#endif
			break ;

		case ENGINE_$INFO:
			DLR_Info() ;
			break ;

		case ENGINE_$GAME:
			DLR_Game() ;
			break ;

		case ENGINE_$LOG:
			DLR_Log() ;
			break ;

		case ENGINE_$MATCH:
			DLR_Match() ;
			break ;

		case  ENGINE_$PDN:
			DLR_PDN() ;
			break ;

		case ENGINE_$PRM:
			DLR_Prm() ;
			break ;

		// $PP - Commands
		case ENGINE_$PPMODE:
			PPMode(true, 0) ;
			break ;

		case ENGINE_$PPPROC:
			PPProc() ;
			break ;

		// !-Commands
		case ENGINE_ASX_DB64:
			ASX_DB64() ;
			break ;

		// #-Commands
		case ENGINE_NMBRSGN_MOVEGENERATE:
			NMBRSGN_MoveGenerate() ;
			break ;

		case ENGINE_NMBRSGN_PERFT:
			NMBRSGN_Perft() ;
			break ;

		case ENGINE_NMBRSGN_TIME:
			NMBRSGN_Time() ;
			break ;

		// @-Commands
		case ENGINE_AT_RUN:
			At_Run() ;
			break ;

		default:
			break ;
	}
}


bool CEngine::DecodeBlitzGameInfo()
{
	CString csInfo, csToken ;

	csInfo = m_EngineMessage[m_iNMessage].csMessageBody ;

	if ( ParseMessage(&csInfo, &csToken) )	// Info string must contain Game Move Nr.
		m_EngineData.iBlitzMove = atoi( LPCTSTR(csToken) ) ;
	else
		return false ;

	if ( ParseMessage(&csInfo, &csToken) )	// Info string must contain Blitz Time (sec.)
		m_EngineData.iBlitzTimeLeft = m_EngineData.iBlitzTime - atoi(LPCTSTR(csToken) ) ;
	else
		return false ;

	return true ;
}


void CEngine::DecodeEngineMove(int iSearchThreadID, int iMoveIndex)
{
	int i ;
	PLOOK pMoveStack ;
	PTSTACK pTreeStack ;

	pMoveStack = m_EngineData.SearchThreadData[iSearchThreadID].pMoveStack ;
	pTreeStack = m_EngineData.SearchThreadData[iSearchThreadID].pTreeStack ;

	if ( m_EngineData.iEngineID == ENGINE_ID_DAMAGE || m_EngineData.iEngineID == ENGINE_ID_HORIZON4 ) {
		m_pSearch64->Decode_Move(iMoveIndex, m_iMoveData, pTreeStack, pMoveStack) ;

		for (i=0; i<23; i++ )
			m_EngineData.iMoveData[i] = m_iMoveData[i] ;
	}
	
	if ( m_EngineData.iEngineID == ENGINE_ID_HORIZON3 ) {

		for (i=0; i<23; i++ )
			m_iMoveData[i] = m_EngineData.iMoveData[i] ;
	}
}


int CEngine::DecodeDXPMove(void)
{
	int i, iStart ;
	CString csDXPMove, csToken ;

	csDXPMove = m_EngineMessage[m_iNMessage].csMessageBody ;
	
	csToken			= csDXPMove.Mid(5,2) ;
	m_iMoveData[0]	= atoi( LPCTSTR(csToken) ) ;

	csToken			= csDXPMove.Mid(7,2) ;
	m_iMoveData[1]	= atoi( LPCTSTR(csToken) ) ;

	csToken			= csDXPMove.Mid(9,2) ;	
	m_iMoveData[2]	= atoi( LPCTSTR(csToken) ) ;

	if ( m_iMoveData[2] == 0 )
		m_iMoveData[2] = -1 ;

	if ( m_iMoveData[2] > 0 ) {

		iStart = 11 ;

		for ( i=0 ; i < m_iMoveData[2] ; i++ ) {

			csToken				= csDXPMove.Mid(iStart,2) ;	
			m_iMoveData[i + 3]	= atoi( LPCTSTR(csToken) ) ;

			iStart += 2 ;
		}
	}
		
	for ( i=0; i<23; i++ )
		m_EngineData.iMoveData[i] = m_iMoveData[i] ;

	m_pSearch64->CopyBoard(&m_EngineData) ;

	return ( m_pSearch64->Find_Move(m_iMoveData, m_EngineData.pTreeStack, m_EngineData.pMoveStack) ) ;
}


int CEngine::DecodeGUIMove(void)
{
	int i, iFindMove ;
	CString csMove, csToken ;

	csMove = m_EngineMessage[m_iNMessage].csMessageBody ;

	if ( ParseMessage(&csMove, &csToken) )	// Move string must contain From Position
		m_iMoveData[0] = atoi( LPCTSTR(csToken) ) ;
	else
		return false ;

	if ( ParseMessage(&csMove, &csToken) )	// Move string must contain To Position
		m_iMoveData[1] = atoi( LPCTSTR(csToken) ) ;
	else
		return false ;

	if ( ParseMessage(&csMove, &csToken) )	
		m_iMoveData[2] = atoi( LPCTSTR(csToken) ) ;
	else								// If Move String des not contain NKill value
		m_iMoveData[2] = -1 ;			// insert -1 as value !

	if ( m_iMoveData[2] > 0 ) 
		for ( i=0 ; i < m_iMoveData[2] ; i++ ) {
			ParseMessage(&csMove, &csToken) ;
			m_iMoveData[i + 3] = atoi( LPCTSTR(csToken) ) ;
		}	

	for ( i=0; i<23; i++ )
		m_EngineData.iMoveData[i] = m_iMoveData[i] ;

	m_pSearch64->CopyBoard(&m_EngineData) ;

	iFindMove = m_pSearch64->Find_Move(m_iMoveData, m_EngineData.pTreeStack, m_EngineData.pMoveStack) ;

	if ( iFindMove != -1 ) {

		m_pSearch64->Decode_Move(iFindMove, m_iMoveData, m_EngineData.pTreeStack, m_EngineData.pMoveStack) ;

		for (i=0; i<23; i++ )
			m_EngineData.iMoveData[i] = m_iMoveData[i] ;
	}

	return ( iFindMove ) ;
}


void CEngine::ProcessMove(int *iBoard, int *iMoveData)
{
	int iPFrom, iPTo, iTemp ;
	int i, iNCapture, iPCapture ;

	iPFrom		= iMoveData[0] ;
	iPTo		= iMoveData[1] ;
	iNCapture	= iMoveData[2] ;

	iTemp			= iBoard[iPFrom];
	iBoard[iPFrom]	= 0;
	iBoard[iPTo]	= iTemp;

	if ( iTemp == ENGINE_WHITEMAN && iPTo <= 5 )		// Promotion WhiteMan
		iBoard[iPTo] = ENGINE_WHITEKING;

	if ( iTemp == ENGINE_BLACKMAN && iPTo >= 46 )	// Promotion BlackMan
		iBoard[iPTo] = ENGINE_BLACKKING ;

	if ( iNCapture > 0 ) {

		for ( i=0; i<iNCapture; i++ ) {

			iPCapture = iMoveData[3 + i] ;
			iBoard[iPCapture] = 0 ;
		}
	}
}


void CEngine::FormatMove(int *iMoveData, CString* pcsMove )
{
	int iPFrom, iPTo, iNCapture ;

	iPFrom		= iMoveData[0] ;
	iPTo		= iMoveData[1] ;
	iNCapture	= iMoveData[2] ;

	if ( iNCapture <= 0 )
		pcsMove->Format("%d-%d", iPFrom, iPTo) ;
	else
		pcsMove->Format("%dx%d", iPFrom, iPTo) ;

}


bool CEngine::EngineMode(void)
{
	CString csEngineMode, csMessage, csToken ;
	int iEngineMode, iMove, iSearchValue, iThreshold ;
	
	csEngineMode = m_EngineMessage[m_iNMessage].csMessageBody ;

	if ( ParseMessage(&csEngineMode, &csToken) )	// EngineMode string must contain ENGINE_MODE_XXX
		iEngineMode = atoi( LPCTSTR(csToken) ) ;
	else
		return false ;

	switch ( iEngineMode ) {

		case ENGINE_MODE_NORMAL:
			
			m_EngineData.iEngineMode				= ENGINE_MODE_NORMAL ;
			m_EngineData.iEngineModeMask			= ENGINE_MODEMASK_POSITION ;

			if ( ParseMessage(&csEngineMode, &csToken) )	// NormalMode string can (!) contain auto-move information
				iMove = atoi( LPCTSTR(csToken) ) ;
			else
				iMove = 1 ;

			if ( iMove == 1 )
				m_EngineData.iEngineModeMask		|= ENGINE_MODEMASK_MOVE ;

			break ;

		case ENGINE_MODE_ANALYSIS:

			m_EngineData.iEngineMode				= ENGINE_MODE_ANALYSIS ;
			m_EngineData.iEngineModeMask			= ENGINE_MODEMASK_ANALYSIS ;
			m_EngineData.iAnalysisSearchMode		= ENGINE_SEARCH_FIXED_TIME ;

			if ( ParseMessage(&csEngineMode, &csToken) )	// AnalysisMode string must contain Threshold
				iThreshold = atoi( LPCTSTR(csToken) ) ;
			else
				return false ;

			m_EngineData.iAnalysisThreshold		= iThreshold ;
		
			if ( ParseMessage(&csEngineMode, &csToken) )	// AnalysisMode string must contain SearchValue
				iSearchValue = atoi( LPCTSTR(csToken) ) ;
			else
				return false ;
		
			m_EngineData.iAnalysisSearchValue	= iSearchValue ;
			break ;

		case ENGINE_MODE_GAME:
			m_EngineData.iEngineMode				= ENGINE_MODE_GAME ;
			m_EngineData.iEngineModeMask			= ( ENGINE_MODEMASK_POSITION  | ENGINE_MODEMASK_GAME ) ;

			if ( ParseMessage(&csEngineMode, &csToken) )	// GameMode string can (!) contain auto-move information
				iMove = atoi( LPCTSTR(csToken) ) ;
			else
				iMove = 1 ;

			if ( iMove == 1 )
				m_EngineData.iEngineModeMask		|= ENGINE_MODEMASK_MOVE ;

			break ;

		case ENGINE_MODE_MATCH:
			m_EngineData.iEngineMode				= ENGINE_MODE_MATCH ;
			m_EngineData.iEngineModeMask			= ( ENGINE_MODEMASK_POSITION | ENGINE_MODEMASK_GAME | ENGINE_MODEMASK_MATCH ) ;
			break ;

		case ENGINE_MODE_LEARN:
			m_EngineData.iEngineMode				= ENGINE_MODE_LEARN ;
			m_EngineData.iEngineModeMask			= ENGINE_MODEMASK_LEARN ;
			break ;

		case ENGINE_MODE_INFINITE_ANALYSIS:
			m_EngineData.iEngineMode				= ENGINE_MODE_INFINITE_ANALYSIS ;
			m_EngineData.iEngineModeMask			= ENGINE_MODEMASK_POSITION ;
			break ;
	}

	if ( m_EngineData.bEvalOutput )
		m_EngineData.iEngineModeMask |= ENGINE_MODEMASK_OUTPUT_EVAL ;

	SendEngineMessage(ENGINE_CONSOLE, ENGINE_NOTIFICATION_ONUPDATE_ENGINEMODE, csMessage) ;

	return ( true ) ;
}


void CEngine::ExecuteMove(int iIndexMove)
{
	int i ;
	CString csMove ;

	FormatMove(m_EngineData.iMoveData, &csMove ) ;
	ProcessMove(m_EngineData.iBoard, m_EngineData.iMoveData) ;

	m_pSearch64->ConvertBoard(&m_EngineData) ;

	m_EngineData.bTurn	= ( m_EngineData.bTurn ? false : true ) ;

	// Update GameHistory 
	m_EngineData.GameHistory[m_EngineData.iGameHistoryActualMove].iIndexMove = iIndexMove ;

	if ( m_EngineData.iGameHistoryActualMove - 1 < GAMEHISTORY_MAXSIZE ) {

		m_EngineData.GameHistory[m_EngineData.iGameHistoryActualMove].csMove = csMove ;

		for ( i=0; i<23; i++ )
			m_EngineData.GameHistory[m_EngineData.iGameHistoryActualMove].iMoveData[i] = m_EngineData.iMoveData[i] ;

		++(m_EngineData.iGameHistoryActualMove) ;

		if ( m_EngineData.iGameHistoryActualMove >= m_EngineData.iGameHistoryMaxMove )
			m_EngineData.iGameHistoryMaxMove = m_EngineData.iGameHistoryActualMove ;

		m_EngineData.GameHistory[m_EngineData.iGameHistoryActualMove].bTurn = m_EngineData.bTurn ;

		for ( i=0 ; i<5 ; i++ )
			m_EngineData.GameHistory[m_EngineData.iGameHistoryActualMove].bbField[i] = m_EngineData.bbField[i] ;
	}

	if ( m_EngineData.iEngineModeMask & ENGINE_MODEMASK_OUTPUT_EVAL )
		EvalOutput() ;
}


void CEngine::Forward()
{
	int i, iForward ;

	iForward = atoi(m_EngineMessage[m_iNMessage].csMessageBody) ;
	
	if ( m_EngineData.iGameHistoryActualMove + iForward <= m_EngineData.iGameHistoryMaxMove ) {
		
		m_EngineData.iGameHistoryActualMove += iForward ;

		m_EngineData.bTurn = m_EngineData.GameHistory[m_EngineData.iGameHistoryActualMove].bTurn ;

		for ( i=0 ; i<5 ; i++ )		// 64Bit
			m_EngineData.bbField[i] = m_EngineData.GameHistory[m_EngineData.iGameHistoryActualMove].bbField[i] ;

		m_pSearch64->DeConvertBoard64(&m_EngineData) ;

		m_EngineMessage[m_iNMessage].bDisplayUpdate	= true ;
	}

}


bool CEngine::Game(void)
{
	bool bTurn ;
	int i, iMoveNumber, iGameMoveData[23] ;
	CString csGameMessage, csToken ;

	csGameMessage = m_EngineMessage[m_iNMessage].csMessageBody ;

	if ( ParseMessage(&csGameMessage, &csToken) ) {

		if ( csToken == ENGINE_STR_TURN_WHITE )
			bTurn = true ;
		else if ( csToken == ENGINE_STR_TURN_BLACK )
			bTurn = false ;
		else
			return ( false ) ;
		
	} else
		return ( false ) ;

	if ( ParseMessage(&csGameMessage, &csToken) ) {

		if ( csToken.GetLength() == 50 )
			StringToPosition(bTurn, &csToken) ;
		else
			return ( false ) ;

	} else
		return ( false ) ;

	if ( ParseMessage(&csGameMessage, &csToken) ) 

		iMoveNumber = atoi( LPCTSTR(csToken) ) ;

	else
		return ( false ) ;

	for ( i=0 ; i<iMoveNumber ; i++ )
		ParseGameMove(&csGameMessage, iGameMoveData) ;

	return ( true ) ;
}


void CEngine::DXPGo(void)
{

	if ( m_bDXPGame )
		Go() ;

	if ( m_bDXPMatch )
		MatchGo() ;
}


void CEngine::MatchGo(void)
{
	CString csFileName ;

	m_EngineData.iEngineMode		= ENGINE_MODE_MATCH ;
	m_EngineData.iEngineModeMask	= ( ENGINE_MODEMASK_POSITION | ENGINE_MODEMASK_GAME | ENGINE_MODEMASK_MATCH | ENGINE_MODEMASK_MOVE ) ;

	if ( m_EngineData.bEvalOutput )
		m_EngineData.iEngineModeMask |= ENGINE_MODEMASK_OUTPUT_EVAL ;

	m_EngineData.DXPMatch.iNWin		= 0 ;
	m_EngineData.DXPMatch.iNDraw	= 0 ;
	m_EngineData.DXPMatch.iNLoose	= 0 ;

	// Init Match File
	csFileName = m_EngineData.DXPMatch.csMatchName + ".pdn";

	m_EngineData.DXPMatch.csMatchFile	= csFileName ;
	m_EngineData.DXPMatch.csClientName	= m_EngineData.csEngineName ;

	if ( m_EngineData.DXPMatch.MatchFile.Open(csFileName, CFile::modeWrite) )							// Open existing File
		m_EngineData.DXPMatch.bFileOpen = true ;
	
	else if ( m_EngineData.DXPMatch.MatchFile.Open(csFileName, CFile::modeWrite | CFile::modeCreate ) )	// Create File
		m_EngineData.DXPMatch.bFileOpen = true ;

	else
		m_EngineData.DXPMatch.bFileOpen = false ;

	// Init Match Log File
	csFileName = m_EngineData.DXPMatch.csMatchName + ".log";

	m_EngineData.DXPMatch.csMatchLogFile	= csFileName ;

	DamageLog(DXP_MATCH_LOG_OPEN) ;

	ResetEvent(m_hEventGameStop) ;
	SetEvent(m_hEventMatchStart) ;	// Start Match
}


void CEngine::Go(void)
{
	int i ;
	CString csMessage ;

	m_pSearch64->CopyBoard(&m_EngineData) ;

	if ( m_EngineData.iEngineModeMask & ENGINE_MODEMASK_ANALYSIS )
		m_EngineData.am_idx = m_EngineData.GameHistory[m_EngineData.iGameHistoryActualMove].iIndexMove ;

	if ( ( m_EngineData.iEngineModeMask & ENGINE_MODEMASK_POSITION ) && m_EngineData.iNormalSearchMode == ENGINE_SEARCH_BLITZGAME ) {
		DecodeBlitzGameInfo() ;
	}

	// Init PP-Parameters
	m_EngineData.bPPAbortSearch = false ;
	
	for ( i=0; i<m_EngineData.iPPProcessActive; i++ )
		m_EngineData.SearchThread[i].bStop = false ;

	if ( m_EngineData.bTurn ) 
		m_EngineData.bComp = m_EngineData.bCWhite ;
	else
		m_EngineData.bComp = m_EngineData.bCBlack ;

	if ( !m_EngineData.bComp && ( m_EngineData.iEngineModeMask & ENGINE_MODEMASK_POSITION ) && m_EngineData.bBrain )
		m_EngineData.bThink = true ;
	else
		m_EngineData.bThink = false ;

	if ( m_EngineData.iEngineMode == ENGINE_MODE_BOOK_UPDATE )
		m_EngineData.bThink = true ;

	if ( ( m_EngineData.iEngineModeMask & ENGINE_MODEMASK_POSITION ) && !m_EngineData.bThink ) {
	
		m_EngineData.iSearchMode = m_EngineData.iNormalSearchMode ;

		if ( m_EngineData.iSearchMode == ENGINE_SEARCH_BLITZGAME )
			m_EngineData.iSearchValue = m_pSearch64->Search_Time(&m_EngineData) ;
		else
			m_EngineData.iSearchValue = m_EngineData.iNormalSearchValue ;

	} else if ( ( m_EngineData.iEngineModeMask & ENGINE_MODEMASK_POSITION ) && m_EngineData.bThink ) {
		
		m_EngineData.iSearchMode	= ENGINE_SEARCH_FIXED_TIME ;
		m_EngineData.iSearchValue	= ENGINE_INFINITE_TIME ;
		
	} else if ( m_EngineData.iEngineModeMask & ENGINE_MODEMASK_ANALYSIS ) {

		m_EngineData.iSearchMode	= m_EngineData.iAnalysisSearchMode ;
		m_EngineData.iSearchValue	= m_EngineData.iAnalysisSearchValue ;

	} else if ( m_EngineData.iEngineModeMask & ENGINE_MODEMASK_INFINITE_ANALYSIS ) {

		m_EngineData.iSearchMode	= ENGINE_SEARCH_FIXED_DEPTH ;
		m_EngineData.iSearchValue	= ENGINE_SEARCH_MAXDEPTH ;
	}

	if ( m_EngineData.iSearchMode != ENGINE_SEARCH_FIXED_DEPTH )
		m_EngineData.dTimeSearch = GetTickCount() ;

	if ( m_EngineData.iEngineModeMask & ENGINE_MODEMASK_MOVE )
		m_EngineData.bMove = ( m_EngineData.bThink ? false : true ) ;
	else
		m_EngineData.bMove = false  ;

	if ( m_EngineData.iEngineModeMask & ENGINE_MODEMASK_OUTPUT_EVAL )
		EvalOutput() ;

	if ( m_EngineData.bComp || m_EngineData.bThink || ( m_EngineData.iEngineModeMask & ( ENGINE_MODEMASK_INFINITE_ANALYSIS | ENGINE_MODEMASK_ANALYSIS ) ) ) {

		m_pSearch64->SearchPP_Clear(&m_EngineData) ;

		if ( m_bDebug ) {
			csMessage.Format("SearchMode, SearchValue : %d %d", m_EngineData.iSearchMode, m_EngineData.iSearchValue ) ;
			SendEngineMessage(ENGINE_CONSOLE, ENGINE_$DEBUG, csMessage) ;
		}

		if ( m_EngineData.iEngineMode == ENGINE_MODE_MATCH ) {		// Write info to Match LogFile

			DamageLog(DXP_MATCH_LOG_GAME_BOARD) ;
			DamageLog(DXP_MATCH_LOG_GAME_MOVE) ;
		}

		ThreadStart() ;
	}
}


void CEngine::Halt()
{
	int iValue ;

	iValue = atoi(m_EngineMessage[m_iNMessage].csMessageBody) ;

	if ( iValue  == 1 )
		ThreadStop(true, 0, true) ;
	else
		ThreadStop(true, 0, false) ;

}


void CEngine::Halt(int iHalt)
{
	if ( m_EngineData.iEngineModeMask & ENGINE_MODEMASK_ANALYSIS )

		ThreadStop(true, 0, false) ;		// ENGINE_MODE_ANALYSIS
	else

		ThreadStop(true, 0 , true) ;

}


void CEngine::JumpTo(void)
{
	int i, iJumpTo ;

	iJumpTo = atoi(m_EngineMessage[m_iNMessage].csMessageBody) ;
	
	if ( iJumpTo <= m_EngineData.iGameHistoryMaxMove ) {
		
		m_EngineData.iGameHistoryActualMove = iJumpTo ;

		m_EngineData.bTurn = m_EngineData.GameHistory[m_EngineData.iGameHistoryActualMove].bTurn ;

		for ( i=0 ; i<5 ; i++ )		// 64Bit
			m_EngineData.bbField[i] = m_EngineData.GameHistory[m_EngineData.iGameHistoryActualMove].bbField[i] ;

		m_pSearch64->DeConvertBoard64(&m_EngineData) ;

		m_EngineMessage[m_iNMessage].bDisplayUpdate	= true ;
	}
}


bool CEngine::JumpTo(int iJumpTo)
{
	int i ;
	
	if ( iJumpTo > m_EngineData.iGameHistoryMaxMove )
		return ( false ) ;
		
	m_EngineData.iGameHistoryActualMove = iJumpTo ;

	m_EngineData.bTurn = m_EngineData.GameHistory[m_EngineData.iGameHistoryActualMove].bTurn ;

	for ( i=0 ; i<5 ; i++ )			// 64Bit
		m_EngineData.bbField[i] = m_EngineData.GameHistory[m_EngineData.iGameHistoryActualMove].bbField[i] ;

	m_pSearch64->DeConvertBoard64(&m_EngineData) ;

	m_EngineMessage[m_iNMessage].bDisplayUpdate	= true ;

	return ( true ) ;
}


void CEngine::LevelsBlitzGame(void)
{
	CString csMessage, csToken ;

	csMessage = m_EngineMessage[m_iNMessage].csMessageBody ;

	ParseMessage(&csMessage, &csToken) ;
	m_EngineData.iBlitzTime			= atoi( LPCTSTR(csToken) ) ;
	
	ParseMessage(&csMessage, &csToken) ;
	m_EngineData.iBlitzGameMoves	= atoi( LPCTSTR(csToken) ) ;

	m_EngineData.iBlitzTimeLeft				= m_EngineData.iBlitzTime ;
	m_EngineData.iEngineTimeLeft			= m_EngineData.iBlitzTime ;
	m_EngineData.iBlitzGameWhiteTimeUsed	= 0 ;
	m_EngineData.iBlitzGameBlackTimeUsed	= 0 ;

	m_EngineData.iBlitzGameWhiteMovesLeft	= m_EngineData.iBlitzGameMoves ;
	m_EngineData.iBlitzGameBlackMovesLeft	= m_EngineData.iBlitzGameMoves ;		

	m_EngineData.iNormalSearchMode	= ENGINE_SEARCH_BLITZGAME ;

}


void CEngine::LevelsLongGame()
{
	CString csMessage ;

	csMessage.Format("%s Engine Error\r\nOption LevelsLongGame not implemented in %s", ENGINE_STR_ERROR, m_EngineData.csEngineName ) ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_ERROR, csMessage) ;

}


void CEngine::LevelsFixedDepth(void)
{
	int iValue ;

	iValue = atoi(m_EngineMessage[m_iNMessage].csMessageBody) ;

	m_EngineData.iNormalSearchMode	= ENGINE_SEARCH_FIXED_DEPTH ;
	m_EngineData.iNormalSearchValue	= iValue ;

}


void CEngine::LevelsFixedTime(void)
{
	int iValue ;
	
	iValue = atoi(m_EngineMessage[m_iNMessage].csMessageBody) ;

	m_EngineData.iNormalSearchMode	= ENGINE_SEARCH_FIXED_TIME ;
	m_EngineData.iNormalSearchValue	= iValue ;

}


void CEngine::LevelsInfiniteAnalysis(void)
{
	CString csMessage ;

	m_EngineData.iNormalSearchMode	= ENGINE_SEARCH_FIXED_DEPTH ;
	m_EngineData.iNormalSearchValue	= ENGINE_SEARCH_MAXDEPTH ;
	m_EngineData.iEngineMode		= ENGINE_MODE_INFINITE_ANALYSIS ;
	m_EngineData.iEngineModeMask	= ( ENGINE_MODEMASK_INFINITE_ANALYSIS | ENGINE_MODEMASK_POSITION ) ;

	SendEngineMessage(ENGINE_CONSOLE, ENGINE_NOTIFICATION_ONUPDATE_ENGINEMODE, csMessage) ;

}


void CEngine::DXPMove(void)
{	
	int iIndexDecode ;
	CString csMessage ;

	ThreadStop(true, 0, false) ;

	// In Future add new member parameter iCommunicationMode 0, 1 , 2
	// Com 0:	Stand Alone
	// Com 1:	Guide
	// Com 2:	DXP

	iIndexDecode = DecodeDXPMove() ;	
//	iIndexDecode = DecodeGUIMove() ;

	if ( iIndexDecode != -1 ) {
	
		m_EngineData.bm_idx	= iIndexDecode ;

		ExecuteMove(m_EngineData.bm_idx) ;

		m_EngineMessage[m_iNMessage].bDisplayUpdate	= true ;

	} else {					// Error Decode GUIMove !

		csMessage.Format("%s Engine Error\r\nIllegal GUIMove !", ENGINE_STR_ERROR) ;
		SendEngineMessage(ENGINE_CONSOLE, ENGINE_ERROR, csMessage) ;
	}
}


void CEngine::Move(void)
{	
	int iIndexDecode ;
	CString csMessage ;

	ThreadStop(true, 0, false) ;
	
	iIndexDecode = DecodeGUIMove() ;

	if ( iIndexDecode != -1 ) {
	
		m_EngineData.bm_idx	= iIndexDecode ;

		ExecuteMove(m_EngineData.bm_idx) ;

		m_EngineMessage[m_iNMessage].bDisplayUpdate	= true ;

	} else {					// Error Decode GUIMove !

		csMessage.Format("%s Engine Error\r\nIllegal GUIMove !", ENGINE_STR_ERROR) ;
		SendEngineMessage(ENGINE_CONSOLE, ENGINE_ERROR, csMessage) ;
	}
}


void CEngine::New(void)
{
	int i ;

	ThreadStop(true, 0, false) ;

	for ( i=0 ; i<SEARCHTHREAD_MAXTHREAD ; i++ )
		m_EngineData.bPPSearch[i] = false ;
	
	m_EngineData.bThink		= false ;
	m_EngineData.bCWhite	= false ;
	m_EngineData.bCBlack	= false ;
	m_EngineData.bComp		= false ;
	
	BoardInit() ;	
	
	// Reset Game History
	m_EngineData.iGameHistoryActualMove	= 0 ;
	m_EngineData.iGameHistoryMaxMove	= 0 ;

	// Set Game History Entry 0
	m_EngineData.GameHistory[0].bTurn = m_EngineData.bTurn ;

	for ( i=0 ; i<5 ; i++ )		// 64Bit
		m_EngineData.GameHistory[0].bbField[i] = m_EngineData.bbField[i] ;

	m_EngineMessage[m_iNMessage].bDisplayUpdate	= true ;

	if ( m_EngineData.iEngineModeMask & ENGINE_MODEMASK_OUTPUT_EVAL )
		EvalOutput() ;

	// Set EngineSearch Parameters
	m_EngineData.iBlitzGameWhiteTimeUsed	= 0 ;
	m_EngineData.iBlitzGameBlackTimeUsed	= 0 ;
	m_EngineData.iBlitzGameWhiteMovesLeft	= ( m_EngineData.iSearchMode == ENGINE_SEARCH_BLITZGAME ? m_EngineData.iBlitzGameMoves : ENGINE_MAX_GAME_MOVES ) ;
	m_EngineData.iBlitzGameBlackMovesLeft	= ( m_EngineData.iSearchMode == ENGINE_SEARCH_BLITZGAME ? m_EngineData.iBlitzGameMoves : ENGINE_MAX_GAME_MOVES ) ;
}


void CEngine::OfferDraw(void)
{
	CString csMessage ;

	csMessage.Format("%s Engine Error\r\nOption OfferDraw not implemented in %s", ENGINE_STR_ERROR, m_EngineData.csEngineName) ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_ERROR, csMessage) ;

}


void CEngine::Opponents(void)
{
	CString csOpponents ;

	csOpponents = m_EngineMessage[m_iNMessage].csMessageBody ;

	if ( csOpponents == ENGINE_STR_OPPONENTMODE_PLAYER_PLAYER ) {
		m_EngineData.iOpponents	= ENGINE_OPPONENTMODE_PLAYER_PLAYER ;
		m_EngineData.bCWhite		= false ;
		m_EngineData.bCBlack		= false ;
			
	} else if ( csOpponents == ENGINE_STR_OPPONENTMODE_PLAYER_ENGINE ) {
		m_EngineData.iOpponents	= ENGINE_OPPONENTMODE_PLAYER_ENGINE ;
		m_EngineData.bCWhite		= false ;
		m_EngineData.bCBlack		= true ;
			
	} else if ( csOpponents == ENGINE_STR_OPPONENTMODE_ENGINE_PLAYER ) {
		m_EngineData.iOpponents	= ENGINE_OPPONENTMODE_ENGINE_PLAYER ;
		m_EngineData.bCWhite		= true ;
		m_EngineData.bCBlack		= false ;
		
	} else if ( csOpponents == ENGINE_STR_OPPONENTMODE_ENGINE_ENGINE ) {
		m_EngineData.iOpponents	= ENGINE_OPPONENTMODE_ENGINE_ENGINE ;
		m_EngineData.bCWhite		= true ;
		m_EngineData.bCBlack		= true ;

	}
}


void CEngine::Ping(void)
{
	CString csPingMessage ;

	csPingMessage = m_EngineMessage[m_iNMessage].csMessageBody ;

	Pong(csPingMessage) ;
}


void CEngine::Pong(CString csPongMessage)
{
	CString csMessage ;

	csMessage.Format("%s %s", ENGINE_STR_PONG, csPongMessage ) ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_PONG, csMessage) ;
}


bool CEngine::Position()
{
	bool bTurn, bBoardClear ;
	int i ;
	CString csPositionMessage, csToken ;


	ThreadStop(true, 0, false) ;

	bBoardClear = true ;
	
	csPositionMessage = m_EngineMessage[m_iNMessage].csMessageBody ;

	if ( ParseMessage(&csPositionMessage, &csToken) ) {

		if ( csToken == ENGINE_STR_TURN_WHITE )
			bTurn = true ;
		else if ( csToken == ENGINE_STR_TURN_BLACK )
			bTurn = false ;
		else if ( csToken == ENGINE_STR_TURN_ADD ) {
			bTurn = m_EngineData.bTurn ;
			bBoardClear = false ;
		} else
			return false ;
		
	} else
		return false ;

	if ( bBoardClear) 
		BoardClear() ;

	StringToPosition(bTurn, &csPositionMessage) ;

	m_pSearch64->ConvertBoard(&m_EngineData) ;

	// Reset Game History
	m_EngineData.iGameHistoryActualMove	= 0 ;
	m_EngineData.iGameHistoryMaxMove	= 0 ;
	
	m_EngineData.GameHistory[0].bTurn = m_EngineData.bTurn ;

	for ( i=0 ; i<5 ; i++ )		// 64Bit
		m_EngineData.GameHistory[0].bbField[i] = m_EngineData.bbField[i] ;	

	m_EngineMessage[m_iNMessage].bDisplayUpdate	= true ;

	return true ;
}


void CEngine::RequestGame(void)
{
}


void CEngine::RequestPosition(void)
{
	CString csMessage, csMoveList, csPosition, csSideToMove ;

	csSideToMove = ( m_EngineData.bTurn ? ENGINE_STR_TURN_WHITE : ENGINE_STR_TURN_BLACK ) ;

	PositionToString(&csPosition) ;
	MoveListToString(&csMoveList) ;

	csMessage.Format("%s %s %s%s", ENGINE_STR_POSITION, csSideToMove, csPosition, csMoveList) ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_POSITION, csMessage) ;
}


void CEngine::SetComment(void)
{
}


bool CEngine::SetOption(void)
{
	int iOption, iValue ;
	CString csMessage, csToken ;

	csMessage = m_EngineMessage[m_iNMessage].csMessageBody ;
	
	if ( !ParseMessage(&csMessage, &csToken) )
		return ( false ) ;

	iOption = atoi( LPCTSTR(csToken) ) ;

	if ( !ParseMessage(&csMessage, &csToken) )
		return ( false ) ;

	iValue = atoi( LPCTSTR(csToken) ) ;

	switch ( iOption ) {

		case ENGINE_OPTION_OPENINGBOOK:
			m_EngineData.bBook = ( iValue  ? true : false ) ;
			break ;

		case ENGINE_OPTION_ENDGAMEDATABASE:
			m_EngineData.bDBEnd = ( iValue  ? true : false ) ;
			break ;

		case ENGINE_OPTION_PERMANENTBRAIN:
			m_EngineData.bBrain = ( iValue  ? true : false ) ;
			break ;

		case ENGINE_OPTION_HASHTABLE:
			iValue += HASHTABLESIZE_DEFAULT_NOTABLE ;

			if ( iValue != m_EngineData.iHashTableSize ) {
				m_EngineData.iHashTableSize = iValue ;
				m_pSearch64->Search_HashTable_Modify(&m_EngineData) ;
			}
			break ;

		case ENGINE_OPTION_DBCACHESIZE:

			iValue += DBCACHESIZE_DEFAULT_NOCACHE ;

			if ( iValue == DBCACHESIZE_DEFAULT_NOCACHE )
				m_EngineData.bDBEnd = false ;

			if ( iValue != m_EngineData.iDBCacheSize ) {
				m_EngineData.iDBCacheSize = iValue ;
				m_pSearch64->m_pDatabase64->DB64Cache_Modify(&m_EngineData) ;
			}
			break ;

		case ENGINE_OPTION_DBMAXPIECE:
			if ( iValue > 0 ) ++iValue ;
			m_EngineData.iDBMaxPiece = iValue ;
			break ;

		default:
			break ;

	}

	return ( true ) ;
}


void CEngine::ShowWindow()
{
}


void CEngine::OnOptions(void)
{
	CString csMessage ;

	if ( m_bGuide ) {

		if ( m_EngineData.iEngineID == ENGINE_ID_DAMAGE ) {		// Damage Engine
		
			csMessage.Format("%s %s", ENGINE_STR_NAME, ENGINENAME_D) ;
			SendEngineMessage(ENGINE_CONSOLE, ENGINE_NAME, csMessage ) ;
		
			csMessage.Format("%s %s", ENGINE_STR_AUTHOR, ENGINEAUTHOR_D) ;
			SendEngineMessage(ENGINE_CONSOLE, ENGINE_AUTHOR, csMessage) ;
		
		}

		if ( m_EngineData.iEngineID == ENGINE_ID_HORIZON4 ) {	// Hybride Engine , Damage Search + Horizon Evaluation
		
			csMessage.Format("%s %s", ENGINE_STR_NAME, ENGINENAME_H4) ;
			SendEngineMessage(ENGINE_CONSOLE, ENGINE_NAME, csMessage ) ;
		
			csMessage.Format("%s %s", ENGINE_STR_AUTHOR, ENGINEAUTHOR_H4) ;
			SendEngineMessage(ENGINE_CONSOLE, ENGINE_AUTHOR, csMessage) ;
		
		}

		// Inform GUI about Engine Options 

		csMessage.Format("%s %d Opening_Book %s %d %s %d", 
			ENGINE_STR_OPTION, 
			ENGINE_OPTION_OPENINGBOOK,
			ENGINE_STR_OPTIONS_TYPE, ENGINE_CHECKMARK, 
			ENGINE_STR_OPTIONS_DEFAULT,m_EngineData.bBook ) ;
		SendEngineMessage(ENGINE_CONSOLE, ENGINE_OPTION, csMessage) ;

		csMessage.Format("%s %d EndGame_DataBase %s %d %s %d", 
			ENGINE_STR_OPTION, 
			ENGINE_OPTION_ENDGAMEDATABASE,
			ENGINE_STR_OPTIONS_TYPE, ENGINE_CHECKMARK, 
			ENGINE_STR_OPTIONS_DEFAULT, m_EngineData.bDBEnd ) ;
		SendEngineMessage(ENGINE_CONSOLE, ENGINE_OPTION, csMessage) ;

		csMessage.Format("%s %d Permanent_Brain %s %d %s %d", 
			ENGINE_STR_OPTION,
			ENGINE_OPTION_PERMANENTBRAIN,
			ENGINE_STR_OPTIONS_TYPE, ENGINE_CHECKMARK, 
			ENGINE_STR_OPTIONS_DEFAULT, m_EngineData.bBrain ) ;
		SendEngineMessage(ENGINE_CONSOLE, ENGINE_OPTION, csMessage) ;

		csMessage.Format(	"%s %d Hash_Table %s %d %s %d %s Off %s 16M %s 32M %s 64M %s 128M %s 256M %s 512M", 
			ENGINE_STR_OPTION, 
			ENGINE_OPTION_HASHTABLE,
			ENGINE_STR_OPTIONS_TYPE, ENGINE_COMBO,
			ENGINE_STR_OPTIONS_DEFAULT, m_EngineData.iHashTableSize - HASHTABLESIZE_DEFAULT_NOTABLE,
			ENGINE_STR_OPTIONS_VAR,
			ENGINE_STR_OPTIONS_VAR,
			ENGINE_STR_OPTIONS_VAR,
			ENGINE_STR_OPTIONS_VAR,
			ENGINE_STR_OPTIONS_VAR,
			ENGINE_STR_OPTIONS_VAR,
			ENGINE_STR_OPTIONS_VAR ) ;
		SendEngineMessage(ENGINE_CONSOLE, ENGINE_OPTION, csMessage) ;

		csMessage.Format(	"%s %d DB_Cache_Size %s %d %s %d %s Off %s 256M %s 512M %s 1G %s 2G %s 4G %s 8G", 
			ENGINE_STR_OPTION, 
			ENGINE_OPTION_DBCACHESIZE,
			ENGINE_STR_OPTIONS_TYPE, ENGINE_COMBO,
			ENGINE_STR_OPTIONS_DEFAULT, m_EngineData.iDBCacheSize - DBCACHESIZE_DEFAULT_NOCACHE,
			ENGINE_STR_OPTIONS_VAR,
			ENGINE_STR_OPTIONS_VAR,
			ENGINE_STR_OPTIONS_VAR,
			ENGINE_STR_OPTIONS_VAR,
			ENGINE_STR_OPTIONS_VAR,
			ENGINE_STR_OPTIONS_VAR,
			ENGINE_STR_OPTIONS_VAR ) ;
		SendEngineMessage(ENGINE_CONSOLE, ENGINE_OPTION, csMessage) ;

		csMessage.Format(	"%s %d DB_Max._Piece %s %d %s %d %s Off %s 2 %s 3 %s 4 %s 5 %s 6 %s 7", 
			ENGINE_STR_OPTION, 
			ENGINE_OPTION_DBMAXPIECE,
			ENGINE_STR_OPTIONS_TYPE, ENGINE_COMBO,
			ENGINE_STR_OPTIONS_DEFAULT, m_EngineData.iDBMaxPiece - 1,
			ENGINE_STR_OPTIONS_VAR,
			ENGINE_STR_OPTIONS_VAR,
			ENGINE_STR_OPTIONS_VAR,
			ENGINE_STR_OPTIONS_VAR,
			ENGINE_STR_OPTIONS_VAR,
			ENGINE_STR_OPTIONS_VAR,
			ENGINE_STR_OPTIONS_VAR ) ;
		SendEngineMessage(ENGINE_CONSOLE, ENGINE_OPTION, csMessage) ;

		// Engine Activated
		csMessage.Format("%s", ENGINE_STR_ENGINEOK ) ;
		SendEngineMessage(ENGINE_CONSOLE, ENGINE_ENGINEOK, csMessage) ;

		// ENGINE_STATUS_OK
		m_iEngineStatus = ENGINE_STATUS_OK ;

	}
}


bool CEngine::DLR_ClearCount()
{
	int i,j ;

	for ( i=0 ; i < SEARCHTHREAD_MAXTHREAD ; i++ )
		for ( j=0 ; j<=63 ; j++ )
			m_EngineData.SearchThreadData[i].isc[j] = 0 ;

	return ( true ) ;
}


bool CEngine::DLR_ClearOutput()
{
	int iOutput ;
	CString csMessage, csToken ;

	csMessage = m_EngineMessage[m_iNMessage].csMessageBody ;
	
	if ( !ParseMessage(&csMessage, &csToken) )
		return false ;

	iOutput = atoi( LPCTSTR(csToken) ) ;

	ClearOutput(iOutput) ;

	return ( true ) ;
}


bool CEngine::DLR_Debug()
{
	int i ;
	CString csMessage, csToken ;


	csMessage = m_EngineMessage[m_iNMessage].csMessageBody ;
	
	if ( !ParseMessage(&csMessage, &csToken) ) {

		m_bDebug = ( m_bDebug ? false : true ) ;

		return ( false ) ;
	}

	if ( csToken == "c" )
		m_EngineData.iDebugMode = DLR_DEBUG_OFF ;

	else if ( csToken == "p" ) {

		m_EngineData.iDebugMode = ( DLR_DEBUG_ON | DLR_DEBUG_POSITION ) ;

		m_pSearch64->ConvertBoard(&m_EngineData) ;

		m_EngineData.DebugData.bTurn = m_EngineData.bTurn ;

		for ( i=0; i<4; i++ )
			m_EngineData.DebugData.bbField[i] = m_EngineData.bbField[i] ;

	} else if ( csToken == "d" )
		m_EngineData.iDebugMode |= DLR_DEBUG_DISPLAY ;

	return ( true ) ;
}


void CEngine::DLR_DXP(void)
{
	int iBallot, iMaxMoves, iMaxTime, iPort ;
	CString csIPAddress, csMessage, csToken ;

	iPort		= DXP_PORT ;
	csIPAddress	= DXP_IPADDRESS ;
	
	csMessage = m_EngineMessage[m_iNMessage].csMessageBody ;

	if ( ParseMessage(&csMessage, &csToken) ) {

		if ( csToken == DLR_DXP_STR_CLIENT ) {	// Client

			if ( ParseMessage(&csMessage, &csToken) )
				csIPAddress = csToken ;

			m_iCommunicationMode = ENGINE_COMMUNICATIONMODE_DXP ;
			m_pGuide->Init(GUIDE_CLIENT, m_iCommunicationMode, iPort, csIPAddress) ;
		}

		if ( csToken == DLR_DXP_STR_SERVER ) {	// Server

			m_iCommunicationMode = ENGINE_COMMUNICATIONMODE_DXP ;
			m_pGuide->Init(GUIDE_SERVER, m_iCommunicationMode, iPort, csIPAddress) ;
		}

		if ( csToken == DLR_DXP_STR_BALLOT ) {	// Ballot

			if ( ParseMessage(&csMessage, &csToken) ) {

				iBallot = atoi( LPCTSTR(csToken) ) ;
	
			}
		}

		if ( csToken == DLR_DXP_STR_GO ) {		// Go

			if ( ParseMessage(&csMessage, &csToken) ) {

				iBallot = atoi( LPCTSTR(csToken) ) ;
				m_EngineData.DXPMatch.iNGames = iBallot ;

			} else
				m_EngineData.DXPMatch.iNGames = 0 ;

			if ( m_bDXPGame == true )
				DXPClientStart() ;

			if ( m_bDXPMatch == true )
				MatchGo() ;
		}

		if ( csToken == DLR_DXP_STR_STAT ) {	// Stat

			if ( ParseMessage(&csMessage, &csToken) )
				DXPStatMatchFile(&csToken) ;
			else
				DXPStat() ;
		}

		if ( csToken == DLR_DXP_STR_GAME ) {	// Game
			m_bDXPGame	= true ;
			m_bDXPMatch = false ;
		}

		if ( csToken == DLR_DXP_STR_MATCH ) {	// Match

			if ( ParseMessage(&csMessage, &csToken) ) {

				iBallot = atoi( LPCTSTR(csToken) ) ;
				m_EngineData.DXPMatch.iMatchMode = iBallot ;
			} else
				m_EngineData.DXPMatch.iMatchMode = 0 ;

			DLR_DXP_MatchMode() ;

			m_bDXPGame	= false ;
			m_bDXPMatch = true ;
		}

		if ( csToken == DLR_DXP_STR_PRM ) {		// Prm

			if ( ParseMessage(&csMessage, &csToken) ) {

				if ( csToken == DLR_DXP_STR_PRM_MAXMOVES ) {	// MaxMoves

					if ( ParseMessage(&csMessage, &csToken) ) {

						iMaxMoves = atoi( LPCTSTR(csToken) ) ;
						m_EngineData.DXPMatch.iMaxMoves = iMaxMoves ;
					}
				}

				if ( csToken == DLR_DXP_STR_PRM_MAXTIME ) {		// MaxTime

					if ( ParseMessage(&csMessage, &csToken) ) {

						iMaxTime = atoi( LPCTSTR(csToken) ) ;
						m_EngineData.DXPMatch.iMaxTime = iMaxTime ;
					}
				}

			} else {	// Display DXP_PRM Parameters

				DLR_DXP_Prm_Display() ;

			}
		}
	}
}


bool CEngine::DLR_DXP_MatchMode(void)
{
	bool bTurn ;
	int i ;
	PTSTACK pTDepth ;


	if ( m_EngineData.DXPMatch.iMatchMode == 0 )
		return ( false ) ;

	pTDepth = m_EngineData.pTreeStack ;
	bTurn	= m_EngineData.bTurn ;

	for ( i=0; i<5; i++ ) {
		pTDepth->bbField[i]					= m_EngineData.bbField[i] ;
		m_EngineData.DXPMatch.bbField[i]	= m_EngineData.bbField[i] ;
	}

	pTDepth->bTurn = true ;
	m_pSearch64->m_pMoveGenerate64->MoveGenerate(pTDepth) ;
	m_EngineData.DXPMatch.iWhiteMoves = pTDepth->iXMoves ;

	pTDepth->bTurn						= false  ;
	m_EngineData.DXPMatch.iBlackMoves	= m_pSearch64->m_pMoveGenerate64->MoveGenerate(pTDepth) ;

	m_EngineData.bTurn = bTurn ;

	return ( true ) ;
}


void CEngine::DLR_DXP_Prm_Display(void) 
{
	CString csMessage ;

	// MaxMoves
	csMessage.Format("Max Moves = %d", m_EngineData.DXPMatch.iMaxMoves) ;
	SendEngineMessage(SEARCH_OUTPUT_CONSOLE, ENGINE_ENGINEINFO_DXP, csMessage) ;

	// MaxTime
	csMessage.Format("Max Time = %d", m_EngineData.DXPMatch.iMaxTime) ;
	SendEngineMessage(SEARCH_OUTPUT_CONSOLE, ENGINE_ENGINEINFO_DXP, csMessage) ;

}

#ifdef _ENGINE_DAMAGE
bool CEngine::DLR_Eval(void)
{	
	bool bOutputFile ;
	char cBuf[256] ;
	int i, iEvalOutput, iCount, iCount0, iEval0, iNError, iNEError, iNEvalError, iScore_Eval0, iScore_Eval1 ;
	PTSTACK pTDepth ;
	CString csFileName, csOutputFileName, csMessage, csToken ;
	FILE *fp, *fpOut ;


	csMessage = m_EngineMessage[m_iNMessage].csMessageBody ;
	
	iEvalOutput = DLR_EVAL_OUTPUT ;

	if ( ParseMessage(&csMessage, &csToken) ) {

		if ( csToken == "t" )
			iEvalOutput = DLR_EVAL_TEST ;

		else {

			iEvalOutput = atoi( LPCTSTR(csToken) ) ;

			m_EngineData.bEvalOutput = ( iEvalOutput ? true : false ) ;

			if ( iEvalOutput == DLR_EVAL_NO_OUTPUT )
				if ( m_EngineData.iEngineModeMask & ENGINE_MODEMASK_OUTPUT_EVAL )
					m_EngineData.iEngineModeMask ^= ENGINE_MODEMASK_OUTPUT_EVAL ;

			if ( iEvalOutput == DLR_EVAL_OUTPUT )
				m_EngineData.iEngineModeMask |= ENGINE_MODEMASK_OUTPUT_EVAL ;
		}
	}

	if ( iEvalOutput == DLR_EVAL_OUTPUT )
		EvalOutput() ;

	if ( iEvalOutput == DLR_EVAL_TEST ) {

		csFileName			= m_csScriptsPath + ENGINE_EVAL64_TESTFILE ;
		csOutputFileName	= m_csScriptsPath + ENGINE_EVAL64_OUTPUTTESTFILE ;

		if ( fopen_s(&fp, csFileName, "r") != 0 )			// Open ENGINE_EVAL64_TESTFILE
			return ( false ) ;

		if ( fopen_s(&fpOut, csOutputFileName, "r") == 0 )
			bOutputFile = true ;
		else {

			bOutputFile = false ;

			if ( fopen_s(&fpOut, csOutputFileName, "w") != 0 )	// Open ENGINE_EVAL64_OUTPUTTESTFILE
				return ( false ) ;
		}

		iCount		= 0 ;
		iNError		= 0 ;
		iNEvalError	= 0 ;
		iNEError	= 0 ;

		while ( fgets(cBuf, sizeof(cBuf), fp) ) {

			if ( FENRead(cBuf) == false )
				++iNError ;
			else {

				pTDepth = m_EngineData.SearchThreadData[0].pTreeStack ;

				for ( i=0 ; i<5 ; i++ )		// 64Bit
					pTDepth->bbField[i] = m_EngineData.bbField[i] ;

				pTDepth->bTurn = m_EngineData.bTurn ;

				m_pSearch64->MaterialValue(pTDepth) ;

				iScore_Eval0 = m_pSearch64->SearchEvaluation(m_EngineData.iEngineID, pTDepth->bTurn, pTDepth, -30000, 30000) ;

				if ( pTDepth->bTurn == false )
					iScore_Eval0 = -iScore_Eval0 ;

				// Rotate Board
				m_pSearch64->ConvertBoard(true, &m_EngineData) ;

				for ( i=0 ; i<5 ; i++ )		// 64Bit
					pTDepth->bbField[i] = m_EngineData.bbField[i] ;

				pTDepth->bTurn = ( m_EngineData.bTurn ? false : true ) ;	// Reverse bTurn ( Side to Move )

				m_pSearch64->MaterialValue(pTDepth) ;

				iScore_Eval1 = m_pSearch64->SearchEvaluation(m_EngineData.iEngineID, pTDepth->bTurn, pTDepth, -30000, 30000) ;

				if ( pTDepth->bTurn == false )
					iScore_Eval1 = -iScore_Eval1 ;

				if ( bOutputFile )
					fscanf(fpOut, "%d %d\n", &iCount0, &iEval0) ;
				else
					fprintf(fpOut,"%d %d\n", iCount+1, iScore_Eval1) ;

				if ( iScore_Eval1 != iEval0 )
					++iNEError ;

				if ( iScore_Eval0 != -iScore_Eval1 ) {

					++iNEvalError ;

					// Not Rotate Board
					m_pSearch64->ConvertBoard(false, &m_EngineData) ;

					for ( i=0 ; i<5 ; i++ )		// 64Bit
						pTDepth->bbField[i] = m_EngineData.bbField[i] ;

					pTDepth->bTurn = m_EngineData.bTurn ;

					m_pSearch64->MaterialValue(pTDepth) ;

					iScore_Eval0 = m_pSearch64->SearchEvaluation(m_EngineData.iEngineID, pTDepth->bTurn, pTDepth, -30000, 30000) ;

					// Rotate Board
					m_pSearch64->ConvertBoard(true, &m_EngineData) ;

					for ( i=0 ; i<5 ; i++ )		// 64Bit
						pTDepth->bbField[i] = m_EngineData.bbField[i] ;

					pTDepth->bTurn = ( m_EngineData.bTurn ? false : true ) ;	// Reverse bTurn ( Side to Move )

					m_pSearch64->MaterialValue(pTDepth) ;

					iScore_Eval1 = m_pSearch64->SearchEvaluation(m_EngineData.iEngineID, pTDepth->bTurn, pTDepth, -30000, 30000) ;
				}
			}

			++iCount ;
		}

		fclose(fp) ;
		fclose(fpOut) ;

		csMessage.Format("Count = %d , Error : iN=%d , iNEval=%d , iNEE=%d", iCount, iNError, iNEvalError, iNEError) ;
		SendEngineMessage(ENGINE_CONSOLE, ENGINE_$EVAL, csMessage) ;

	}

	return ( true ) ;
}
#endif


bool CEngine::DLR_Info()
{
	int iInfo ;
	CString csMessage, csToken ;

	csMessage = m_EngineMessage[m_iNMessage].csMessageBody ;
	
	if ( !ParseMessage(&csMessage, &csToken) )
		return false ;

	iInfo = atoi( LPCTSTR(csToken) ) ;

	switch ( iInfo ) {

		case DLR_INFO_PVPOSITION:
			DLR_Info_PVPosition() ;
			break ;

		default:
			break ;
	}

	return ( true ) ;
}


bool CEngine::DLR_Game()
{
	if ( m_EngineData.iEngineModeMask  & ENGINE_MODEMASK_GAME )
		SetEvent(m_hEventGameStart) ;

	return ( true ) ;
}


bool CEngine::DLR_Log()
{
	int i, iLogPly, iNWrite ;
	CString csLogMessage, csMessage, csToken ;
	FILE *pLogFile ;

	csMessage = m_EngineMessage[m_iNMessage].csMessageBody ;
	
	if ( !ParseMessage(&csMessage, &csToken) )
		iLogPly = LOG_PLY_DEFAULT ;
	else
		iLogPly = atoi( LPCTSTR(csToken) ) ;
		
	iNWrite = m_EngineData.iGameHistoryMaxMove ;

	if ( fopen_s(&pLogFile, "LogFile.dat", "w") == 0 ) {

		for ( i=0 ; i<iNWrite ; i++ ) {

			csLogMessage = m_EngineData.GameHistory[i].LogData[iLogPly].csLogMessage ;

			fprintf(pLogFile,"%s\n", csLogMessage) ;
		}

		fclose(pLogFile) ;
	}

	return ( true ) ;
}


bool CEngine::DLR_Match()
{
	int iMessageID, iLog, iNGames ;
	CString csMessage, csMessageBody, csMessageCommand, csToken ;

	csMessage = m_EngineMessage[m_iNMessage].csMessageBody ;

	if ( !CommandMessage(csMessage, &csMessageCommand, &csMessageBody) )
		return ( false ) ;

	iMessageID = -1 ;

	if ( csMessageCommand == DLR_MATCH_STR_ALTERNATE )
		iMessageID = DLR_MATCH_ALTERNATE ;

	else if ( csMessageCommand == DLR_MATCH_STR_BOOK )
		iMessageID = DLR_MATCH_BOOK ;

	else if ( csMessageCommand == DLR_MATCH_STR_GAMES )
		iMessageID = DLR_MATCH_GAMES ;

	else if ( csMessageCommand == DLR_MATCH_STR_GO )
		iMessageID = DLR_MATCH_GO ;

	else if ( csMessageCommand == DLR_MATCH_STR_LOG )
		iMessageID = DLR_MATCH_LOG ;

	else if ( csMessageCommand == DLR_MATCH_STR_NAME )
		iMessageID = DLR_MATCH_NAME ;

	if ( iMessageID == -1 )
		return ( false ) ;

	switch ( iMessageID ) {

		case DLR_MATCH_ALTERNATE:

			m_EngineData.DXPMatch.bAlternate = true ;
			break ;

		case DLR_MATCH_BOOK:
			break ;

		case DLR_MATCH_GAMES:

			if ( !ParseMessage(&csMessage, &csToken) )
				return ( false ) ;

			iNGames = atoi( LPCTSTR(csToken) ) ;
			
			m_EngineData.DXPMatch.iNGames	= iNGames ;
			break ;

		case DLR_MATCH_GO:

			m_EngineData.iEngineMode		= ENGINE_MODE_MATCH ;
			m_EngineData.iEngineModeMask	= ( ENGINE_MODEMASK_POSITION | ENGINE_MODEMASK_GAME | ENGINE_MODEMASK_MATCH | ENGINE_MODEMASK_MOVE ) ;

			m_EngineData.DXPMatch.iNWin	= 0 ;
			m_EngineData.DXPMatch.iNDraw	= 0 ;
			m_EngineData.DXPMatch.iNLoose	= 0 ;

			SetEvent(m_hEventMatchStart) ;	// Start Match
			break ;

		case DLR_MATCH_LOG:

			if ( !ParseMessage(&csMessage, &csToken) )
				return ( false ) ;

			iLog = atoi( LPCTSTR(csToken) ) ;
			
			m_EngineData.DXPMatch.iLog = iLog ;
			break ;

		case DLR_MATCH_NAME:

			m_EngineData.DXPMatch.csMatchName = csMessageBody ;
			break ;

		default:
			break ;
	}

	return ( true ) ;
}


bool CEngine::DLR_PDN(void)
{
	int iMessageID, iNGames, iNMoves, iNPDNGames, iFileRead, iFileSize ;
	CString csMessage, csMessageBody, csMessageCommand, csPDNFileName, csFileName, csToken ;
	FILE* pfPDNFile ;
	GAMEHEADER GameHeader ;

	csMessage = m_EngineMessage[m_iNMessage].csMessageBody ;

	if ( !CommandMessage(csMessage, &csMessageCommand, &csMessageBody) )
		return ( false ) ;

	iMessageID = -1 ;

	if ( csMessageCommand == DLR_PDN_STR_OPEN )
		iMessageID = DLR_PDN_OPEN ;

	else if ( csMessageCommand == DLR_PDN_STR_READ )
		iMessageID = DLR_PDN_READ ;

	else if ( csMessageCommand == DLR_PDN_STR_INFO )
		iMessageID = DLR_PDN_INFO ;

	else if ( csMessageCommand == DLR_PDN_STR_HEADER )
		iMessageID = DLR_PDN_HEADER ;

	if ( iMessageID == -1 )
		return ( false ) ;

	switch ( iMessageID ) {

		case DLR_PDN_OPEN:
			if ( !ParseMessage(&csMessageBody, &csFileName) )
				return ( false ) ;

			csPDNFileName = m_csGamesPath + csFileName ;

			if ( fopen_s(&pfPDNFile, csPDNFileName, "rb" ) != 0 )
				return ( false ) ;

			fseek(pfPDNFile, 0, SEEK_END ) ;		// Calculate FileSize 
			iFileSize = ftell(pfPDNFile) ;
			fseek(pfPDNFile, 0, SEEK_SET) ;

			if ( m_pcPDN != NULL )
				free( m_pcPDN ) ;

			m_pcPDN = (char*)malloc(iFileSize+1) ;

			if ( m_pcPDN == NULL ) {
				fclose(pfPDNFile) ;
				return ( false ) ;
			}

			iFileRead = (int)fread(m_pcPDN, sizeof(char), iFileSize, pfPDNFile) ;

			m_pcPDN[iFileSize] = '\0' ;

			fclose(pfPDNFile) ;

			iNPDNGames = m_pPDN->Read_PDN(m_pcPDN, iFileSize) ;

			csMessage.Format("Games = %d ",iNPDNGames) ;
			SendEngineMessage(ENGINE_CONSOLE, ENGINE_$PDN, csMessage) ;

			break ;

		case DLR_PDN_READ:
			if ( !ParseMessage(&csMessageBody, &csToken) )
				return ( false ) ;

			iNGames = atoi( LPCTSTR(csToken) ) ;

			iNMoves = m_pPDN->Read_Game(iNGames, &GameHeader) ;

			m_iPDNMoves = iNMoves ;

			PDNCopyBoard() ;
			PDNCopyMoves() ;

			csMessage.Format("%s", ENGINE_STR_QMRK_DISPLAYBOARD ) ;
			SendEngineMessage(ENGINE_CONSOLE, ENGINE_QMRK_DISPLAYBOARD, csMessage) ;

			break ;

		case DLR_PDN_INFO:
			break ;

		case DLR_PDN_HEADER:
			break ;
	
		default:
			break ;
	}

	return ( true ) ;
}


bool CEngine::DLR_Prm()
{
	int iParameter, iValue, iValue0 = 0 ;
	CString csMessage, csToken ;

	csMessage = m_EngineMessage[m_iNMessage].csMessageBody ;
	
	if ( !ParseMessage(&csMessage, &csToken) ) {
		DLR_Prm_Display() ;
		return false ;
	}

	iParameter = atoi( LPCTSTR(csToken) ) ;

	if ( iParameter < 0 )
		return ( false ) ;

	if ( !ParseMessage(&csMessage, &csToken) )
		return false ;

	iValue = atoi( LPCTSTR(csToken) ) ;

	switch ( iParameter ) {

		case PRM_SEARCH_MCP_DEPTHREDUCTIONFACTOR:
			m_EngineData.iMCP_DepthReductionFactor = iValue ;
			break ;

		case PRM_SEARCH_MCP_CHILD_SEARCH:
			m_EngineData.iMCP_Child_Search = iValue ;
			break ;

		case PRM_SEARCH_PP_OUTPUT:
			m_EngineData.iPPOutput = iValue ;
			break ;

		case PRM_SEARCH_PVSPLIT_SPLIDEPTH:
			m_EngineData.iPVSplit_SplitDepth = iValue ;
			m_EngineData.iPPMinimumSplitDepth = iValue ;
			break ;

		case PRM_SEARCH_SUBSEARCHMODE:
			m_EngineData.iPPSubSearchMode = iValue ;
			break ;

		case PRM_SEARCH_STARTSEARCHDEPTH:
			m_EngineData.iStartSearchDepth = iValue ;
			break ;

		case PRM_SEARCH_QSEARCH:
			m_EngineData.bQSearch = ( iValue > 0 ? true : false ) ;
			break ;

		case PRM_SEARCH_FHR:
			m_EngineData.bFHR = ( iValue > 0 ? true : false ) ;
			break ;

		case PRM_SEARCH_SNGEX:
			m_EngineData.bSngEx = ( iValue > 0 ? true : false ) ;
			break ;

		case PRM_SEARCH_COUNTCLEAR:
			m_EngineData.bCountClear = ( iValue > 0 ? true : false ) ;
			break ;

		case PRM_SEARCH_LOGGING:
			m_EngineData.bLogging = ( iValue > 0 ? true : false ) ;
			break ;

		case PRM_SEARCH_HASHWRITE:
			m_EngineData.bHashWrite = ( iValue > 0 ? true : false ) ;
			break ;

		case PRM_SEARCH_MINMAX:
			m_EngineData.bMinMax = ( iValue > 0 ? true : false ) ;
			break ;

		case PRM_SEARCH_EVALHASHWRITE:
			m_EngineData.bEvalHashWrite = ( iValue > 0 ? true : false ) ;
			break ;

		case PRM_SEARCH_PVSPLIT_SPLITBRANCH:
			m_EngineData.iPPMinimumSplitBranch = iValue ;
			break ;

		case PRM_SEARCH_HISTORY:
			m_EngineData.bHistory = ( iValue > 0 ? true : false ) ;
			break ;

		case PRM_SEARCH_MOVEGEN_MODE:
			m_EngineData.iMoveGenMode = iValue ;
			break ;

		case PRM_SEARCH_ENGINE:

#ifdef _ENGINE_DAMAGE		// Multiple Engine ID's only for Damage Engine
			iValue0					= m_EngineData.iEngineID ;
			m_EngineData.iEngineID	= iValue ;

			if ( iValue == ENGINE_ID_DAMAGE ) {
				m_EngineData.csEngineName	= ENGINENAME_D ;
				m_EngineData.csEngineAuthor = ENGINEAUTHOR_D ;
			} else if ( iValue == ENGINE_ID_HORIZON4 ) {
				m_EngineData.csEngineName	= ENGINENAME_H4 ;
				m_EngineData.csEngineAuthor = ENGINEAUTHOR_H4 ;
			}

			if ( iValue0 != iValue ) {

				OnOptions() ;

				csMessage.Empty() ;		// Message is empty, only Notification ID to EngineApp()
				SendEngineMessage(ENGINE_CONSOLE, ENGINE_NOTIFICATION_ONUPDATE_ENGINEWINDOWTEXT, csMessage) ;
			}
#endif

			break ;

		case PRM_SEARCH_SP_OUTPUT:
			m_EngineData.iSPOutput = iValue ;
			break ;

		case PRM_SEARCH_DEBUG_OUTPUT:
			m_EngineData.iDebugOutput = iValue ;
			break ;

		case PRM_SEARCH_SINGULAREXTENSIONDEPTH:
			m_EngineData.iSingularExtensionDepth = iValue ;
			break ;

		case PRM_SEARCH_HASHREADETC:
			m_EngineData.bHashReadETC = ( iValue > 0 ? true : false ) ;
			break ;

		case PRM_SEARCH_MCP:
			m_EngineData.bMCP = ( iValue > 0 ? true : false ) ;
			break ;

		case PRM_SEARCH_SELECTIVESEARCH:
			m_EngineData.bSelectiveSearch = ( iValue > 0 ? true : false ) ;
			break ;

		case PRM_SEARCH_LMR:
			m_EngineData.bLMR = ( iValue > 0 ? true : false ) ;
			break ;

		case PRM_SEARCH_IID:
			m_EngineData.bIID = ( iValue > 0 ? true : false ) ;
			break ;

		default:
			break ;
	}

	return ( true ) ;
}


void CEngine::DLR_Prm_Display(void) 
{
	CString csMessage ;

	// FHR
	csMessage.Format("%d FHR = %d", PRM_SEARCH_FHR, m_EngineData.bFHR) ;
	SendEngineMessage(SEARCH_OUTPUT_CONSOLE, ENGINE_ENGINEINFO_PRM, csMessage) ;

	// MCP
	csMessage.Format("%d MCP = %d", PRM_SEARCH_MCP, m_EngineData.bMCP) ;
	SendEngineMessage(SEARCH_OUTPUT_CONSOLE, ENGINE_ENGINEINFO_PRM, csMessage) ;

	// LMR
	csMessage.Format("%d LMR = %d", PRM_SEARCH_LMR, m_EngineData.bLMR) ;
	SendEngineMessage(SEARCH_OUTPUT_CONSOLE, ENGINE_ENGINEINFO_PRM, csMessage) ;

}


void CEngine::DLR_Info_PVPosition(void)
{
	bool bTurn ;
	char cBoard[51] ;
	int i, j, iBoard[51], iDepth, iScore, iType ;
	CString csColor, csMessage ;
	PTSTACK pTDepth ;

	pTDepth = m_EngineData.SearchThreadData[0].pTreeStack ;

	m_pSearch64->DeConvertBoard64(pTDepth->PVPosition.bbField, iBoard) ;

	for ( i=1; i<=50; i++ ) {

		if ( iBoard[i] == ENGINE_WHITEMAN )
			cBoard[i] = 'w' ;

		else if ( iBoard[i] == ENGINE_WHITEKING )
			cBoard[i] = 'W' ;

		else if ( iBoard[i] == ENGINE_BLACKMAN )
			cBoard[i] = 'b' ;

		else if ( iBoard[i] == ENGINE_BLACKKING )
			cBoard[i] = 'B' ;

		else
			cBoard[i] = '.' ;
	}

	csMessage.Format("-----------------------") ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_$INFO, csMessage) ;

	for ( i=0; i<5; i++ ) {

		j = 10 * i ;
	
		csMessage.Format("|   %c   %c   %c   %c   %c |", cBoard[j+1], cBoard[j+2], cBoard[j+3], cBoard[j+4], cBoard[j+5]) ;
		SendEngineMessage(ENGINE_CONSOLE, ENGINE_$INFO, csMessage) ;

		csMessage.Format("| %c   %c   %c   %c   %c   |", cBoard[j+6], cBoard[j+7], cBoard[j+8], cBoard[j+9], cBoard[j+10]) ;
		SendEngineMessage(ENGINE_CONSOLE, ENGINE_$INFO, csMessage) ;
	}

	csMessage.Format("-----------------------") ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_$INFO, csMessage) ;

	bTurn	= pTDepth->PVPosition.bTurn ;
	iDepth	= pTDepth->PVPosition.iDepth ;
	iType	= pTDepth->PVPosition.iType ;
	iScore	= pTDepth->PVPosition.iScore ;

	if ( bTurn == true )
		csColor.Format("White") ;
	else
		csColor.Format("Black") ;

	csMessage.Format("Depth = %d, Color = %s, Score = %d, Type = %d", iDepth, csColor, iScore, iType) ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_$INFO, csMessage) ;

}


bool CEngine::PPMode(bool bMode, int iPPMode0)
{
	int iPPMode ;
	CString csMessage, csToken ;

	if ( bMode == false ) {

		m_EngineData.iPPMode = iPPMode0 ;

		csMessage.Empty() ;

		return ( true ) ;
	}

	csMessage = m_EngineMessage[m_iNMessage].csMessageBody ;
	
	if ( !ParseMessage(&csMessage, &csToken) )
		return false ;

	iPPMode = atoi( LPCTSTR(csToken) ) ;

	if ( iPPMode < 0 || iPPMode > PP_MODE_MAXMODE )
		return ( false ) ;

	m_EngineData.iPPMode = iPPMode ;

	csMessage.Empty() ;

	return ( true ) ;
}


bool CEngine::PPProc(void)
{
	int i, iNProc ;
	CString csMessage, csToken ;

	csMessage = m_EngineMessage[m_iNMessage].csMessageBody ;
	
	if ( !ParseMessage(&csMessage, &csToken) )
		return false ;

	iNProc = atoi( LPCTSTR(csToken) ) ;

	if ( iNProc < 1 || iNProc > m_EngineData.iNCore )
		return ( false ) ;

	m_EngineData.iPPProcessActive = iNProc ;

	for ( i=0; i<iNProc; i++ )
		m_EngineData.iPPProcessStatus[i] = PP_PROCESS_IDLE ;
	
	if ( iNProc < m_EngineData.iNCore )
		for ( i=iNProc; i<m_EngineData.iNCore; i++ )
			m_EngineData.iPPProcessStatus[i] = PP_PROCESS_DEACTIVATED ;

	csMessage.Format("%s", ENGINE_STR_$STATUS) ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_$STATUS, csMessage) ;

	return ( true ) ;
}


void CEngine::SearchThreadStart(int iIDSearchThread)
{

	SetEvent(m_EngineData.hEventStart[iIDSearchThread]) ;

}


void CEngine::ThreadStart(void)
{
	int i ;

	m_EngineData.bPanicTime = false ;

	ftime( &(m_EngineData.EngineSearchStartTime) ) ;

	if ( m_EngineData.iEngineModeMask  & ( ENGINE_MODEMASK_POSITION | ENGINE_MODEMASK_ANALYSIS ) ) {

		switch ( m_EngineData.iPPMode ) {

			case PP_MODE_SINGLECORE:

				m_EngineData.ThreadData[m_EngineData.iPPMasterThread].iThreadTask = THREAD_TASK_SEARCH ;
				SetEvent(m_EngineData.hEventStart[m_EngineData.iPPMasterThread]) ;

				break ;

			case PP_MODE_MM_SINGLECORE:

				m_EngineData.ThreadData[m_EngineData.iPPMasterThread].iThreadTask = THREAD_TASK_MM_SEARCH ;
				SetEvent(m_EngineData.hEventStart[m_EngineData.iPPMasterThread]) ;

				break ;

			case PP_MODE_MULTICORE_COPYSEARCH:

				for ( i=0; i<m_EngineData.iNCore; i++ ) {

					if ( m_EngineData.iPPProcessStatus[i] == PP_PROCESS_IDLE ) {

						m_EngineData.ThreadData[i].iThreadTask = THREAD_TASK_SEARCH ;
						SetEvent(m_EngineData.hEventStart[i]) ;
					}
				}

				break ;

			case PP_MODE_MULTICORE_SMP_SEARCH:

				m_EngineData.bPPIdle					= true ;
				m_EngineData.bPPAllThreadsShouldExit	= false ;
				m_EngineData.bPPAbortSearch				= false ;

				// Start all threads ( Master & Slaves )
				for ( i=0; i<m_EngineData.iPPProcessActive; i++ ) {
					m_EngineData.ThreadData[i].iThreadTask = THREAD_TASK_SMP_SEARCH ;
					SetEvent(m_EngineData.hEventStart[i]) ;
				}

				break ;
		}
	}
}


void CEngine::ThreadStop(bool bStopAll, int iSearchThreadID, bool bMove)
{
	bool bThreadStopTimeInfo = false ;
	int i ;
	float fSearchTime ;
	DWORD dwReturnCode ;
	CString csMessage ;
	time_t iNSeconds, iNMilliSeconds ;
	struct timeb ThreadStopTime ;

	if ( bThreadStopTimeInfo ) {

		ftime(&ThreadStopTime) ;

		iNSeconds		= ThreadStopTime.time		- m_EngineData.EngineSearchStartTime.time ;
		iNMilliSeconds	= ThreadStopTime.millitm	- m_EngineData.EngineSearchStartTime.millitm ;

		fSearchTime  = (float)( ( 1000 * iNSeconds )  + iNMilliSeconds ) ;
		fSearchTime	/= 1000. ;

		csMessage.Format("%s %s %7.2f", SEARCH_STR_SEARCHINFO, SEARCH_STR_SEARCHINFO_STRING, fSearchTime) ; 
		SendEngineMessage(ENGINE_CONSOLE, ENGINE_SEARCHINFO, csMessage) ;
	}

	if ( m_EngineData.bSearch ) {

		m_EngineData.bMove	&= bMove ;
		
		if ( bStopAll ) {

			m_EngineData.bPPAbortSearch = true ;

			for ( i=0; i<m_EngineData.iPPProcessActive; i++ ) {
				m_EngineData.SearchThread[i].bStop	= true ;
				m_EngineData.bPPSearch[i]			= false ;
			}

		} else
				m_EngineData.bPPSearch[iSearchThreadID] = false ;

		dwReturnCode = ::WaitForSingleObject( m_EngineData.hEventSearchStop, INFINITE) ;
	}
}


void CEngine::InitEngine(bool bProfile)
{	
	int i, j, iNCore, iNSearchThreadMax ;
	CString csMessage ;
	SYSTEM_INFO SysInfo ;

	// Init SearchThread Data
	
	GetSystemInfo(&SysInfo) ;
	iNCore				= SysInfo.dwNumberOfProcessors ;
	iNSearchThreadMax	= imin(iNCore, SEARCHTHREAD_MAXTHREAD) ;

	m_EngineData.bProfile			= bProfile ;
	
	m_EngineData.bParallel			= SEARCHTHREAD_PARALLEL ;
	m_EngineData.iNCore				= iNCore ;
	m_EngineData.iNSearchThreadMax	= iNSearchThreadMax ;

	m_EngineData.iSPOutput					= SP_OUTPUT_NONE ;
	m_EngineData.iDebugOutput				= DEBUG_OUTPUT_NONE ;

	m_EngineData.iPPProcessActive			= iNCore ;
	m_EngineData.iPPSubSearchMode			= PP_SUBSEARCHMODE_DEFAULT ;
	m_EngineData.iPPMasterThread			= 0 ;
	m_EngineData.iPPOutput					= PP_OUTPUT_NONE ;

	PPMode(false, PP_MODE_DEFAULT) ;

	m_EngineData.iPPMaxThreadsPerSplitPoint = PP_MAXTHREADSPERSPLITPOINT ;
	m_EngineData.iPPMaxActiveSplitPoints	= PP_MAXACTIVESPLITPOINTS ;
	m_EngineData.iPPMinimumSplitBranch		= PP_MINIMUMSPLITBRANCH ;
	m_EngineData.iPPMinimumSplitDepth		= PP_MINIMUMSPLITDEPTH ;

	for ( i=0; i<SEARCHTHREAD_MAXTHREAD; i++ ) {

		m_EngineData.SearchThread[i].iPPActiveSplitPoints = 0 ;

		for ( j=0; j< PP_MAXACTIVESPLITPOINTS; j++ )
			InitializeCriticalSection( &(m_EngineData.SplitPoint[i][j].SPLock) ) ;
	}

	// HashTable
	m_EngineData.iHashTableSize				= HASHTABLESIZE_DEFAULT ;		// 2^iHashTableSize

	// EndGame DB
	m_EngineData.iDBMaxPiece				= DBMAXPIECE_DEFAULT  ;

#ifdef _ENGINE_DAMAGE
	m_EngineData.iDBCacheSize				= DBCACHESIZE_DEFAULT_DAMAGE ;	// 2^iDBCacheSize
#endif

#ifdef _ENGINE_HORIZON
	m_EngineData.iDBCacheSize				= DBCACHESIZE_DEFAULT_HORIZON ;
	m_EngineData.iDBCacheSize				= DBCACHESIZE_DEFAULT_DAMAGE ;	// 2^iDBCacheSize, temporary ......
#endif

	// Search Parameters
	m_EngineData.iStartSearchDepth			= SEARCH_STARTSEARCHDEPTH ;
	m_EngineData.iPVSplit_SplitDepth		= PP_PVSPLIT_SPLITDEPTH_THRESHOLD ;
	m_EngineData.iMCP_DepthReductionFactor	= SEARCH_MCP_DEPTHREDUCTIONFACTOR ;
	m_EngineData.iMCP_Child_Search			= SEARCH_MCP_CHILD_SEARCH ;
	m_EngineData.iMoveGenMode				= SEARCH_MOVEGEN_MODE ;

	m_pSearch64->m_csTablesPath						= m_csTablesPath ;

	// Copy Database-Path String
	m_pSearch64->m_csDatabasePath					= m_csDatabasePath ;
	m_pSearch64->m_pDatabase64->m_csDatabasePath	= m_csDatabasePath ;

	// Init Search
	m_EngineData.bSharedHashTable		= SEARCHTHREAD_SHARED_HASHTABLE ;
	m_EngineData.bSharedEvalHashTable	= SEARCHTHREAD_SHARED_EVALHASHTABLE ;
	m_EngineData.bSharedHistory			= SEARCHTHREAD_SHARED_HISTORY ;
	m_EngineData.bCountClear			= false ;

	m_pSearch64->Search_Init(&m_EngineData) ;
	m_pSearch64->Search_Init_HTM(&m_EngineData) ;

	m_bDebug				= false ;		// No Debug Mode Initially !
	m_EngineData.iDebugMode	= DLR_DEBUG_OFF ;

	// Init Database
	csMessage.Format("%s Database", ENGINE_STR_ENGINEINFO_INIT) ;
	SendEngineMessage(SEARCH_OUTPUT_CONSOLE, ENGINE_ENGINEINFO_INIT, csMessage) ;

	// Temporary Message
	csMessage.Format("%s %s", ENGINE_STR_STATUS, m_csDatabasePath) ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_STATUS, csMessage) ;

	 m_pSearch64->m_pDatabase64->DB64_Init(&m_EngineData) ;


	// ENGINE_STATUS_INIT
	m_iEngineStatus = ENGINE_STATUS_INIT ;
	m_iGenerateMode = ENGINE_GENERATE_MODE_NONE ;

	m_iNMessage		= 0 ;	// Init/Clear Message Buffer Count

#ifdef _ENGINE_DAMAGE
	m_EngineData.iEngineID			= ENGINE_ID_DAMAGE ;
	m_EngineData.csEngineName		= ENGINENAME_D ;
	m_EngineData.csEngineAuthor		= ENGINEAUTHOR_D ;
#endif

#ifdef _ENGINE_HORIZON
	m_EngineData.iEngineID			= ENGINE_ID_HORIZON4 ;
	m_EngineData.csEngineName		= ENGINENAME_H4 ;
	m_EngineData.csEngineAuthor		= ENGINEAUTHOR_H4 ;
#endif

	m_EngineData.iEngineMode		= ENGINE_MODE_NORMAL ;
	m_EngineData.iEngineModeMask	= ( ENGINE_MODEMASK_POSITION | ENGINE_MODEMASK_MOVE ) ;

	m_EngineData.bBook				= false ;
	m_EngineData.bDBEnd				= true ;
	m_EngineData.bQSearch			= true ;	// Quiescence Search
	m_EngineData.bFHR				= true ;	// Fail High Reduction
	m_EngineData.bHistory			= true ;	// History Table
	m_EngineData.bSngEx				= false ;	// Singular Extensions
	m_EngineData.bHashWrite			= true ;	// Hash Write
	m_EngineData.bMinMax			= false ;	// MinMax Search
	m_EngineData.bEvalHashWrite		= true ;	// Eval Hash Write
	m_EngineData.bHashReadETC		= true ;	// Hash Read ETC
	m_EngineData.bMCP				= true ;	// Search MCP
	m_EngineData.bSelectiveSearch	= false ;	// Selective Search
	m_EngineData.bLMR				= true ;	// Late Move Reductions
	m_EngineData.bIID				= false ;	// Internal Iterative Deepening

	m_EngineData.bLearnDatabase		= false ;
	m_EngineData.bLearnRead			= false ;
	m_EngineData.bLearnWrite		= false ;

	m_EngineData.bBrain				= false ;

	m_EngineData.bLogging			= false ;
	m_EngineData.iLogMode			= LOG_DEFAULT_MODE ;

	m_EngineData.bEvalOutput		= false ;

	m_EngineData.bBook_Active		= false ;
	m_EngineData.bBook_Open			= false ;
	m_EngineData.bBook_Use			= false ;
	m_EngineData.iBookMode			= BOOK_SCORE ;

	m_EngineData.bTurn				= true ;	// White Player Turn

	m_EngineData.bCWhite			= false ;
	m_EngineData.bCBlack			= false ;
	m_EngineData.bComp				= false ;

	for ( i=0; i<SEARCHTHREAD_MAXTHREAD;i++ ) {

		m_EngineData.iPPSearchThreadStatus[i]	= PP_SEARCHTHREAD_IDLE ;
		m_EngineData.bPPSearch[i]				= false ;
	}

	m_EngineData.bThink				= false ;

	// Init Search Parameters
	m_EngineData.iNormalSearchMode			= ENGINE_SEARCH_FIXED_TIME ;
	m_EngineData.iSearchMode				= ENGINE_SEARCH_FIXED_TIME ;

	m_EngineData.iNormalSearchValue			= 10 ;
	m_EngineData.iSearchValue				= 10 ;

	m_EngineData.iBlitzGameWhiteTimeUsed	= 0 ;
	m_EngineData.iBlitzGameBlackTimeUsed	= 0 ;
	m_EngineData.iBlitzGameWhiteMovesLeft	= ENGINE_MAX_GAME_MOVES ;
	m_EngineData.iBlitzGameBlackMovesLeft	= ENGINE_MAX_GAME_MOVES ;	
	
	m_EngineData.iAnalysisSearchMode		= ENGINE_SEARCH_FIXED_TIME ;
	m_EngineData.iAnalysisThreshold			= 30 ;
	m_EngineData.iAnalysisSearchValue		= 5 ;

	m_EngineData.iSingularExtensionDepth	= SINGULAREXTENSIONDEPTH_DEFAULT ;

	m_EngineData.iOpponents					= ENGINE_OPPONENTMODE_PLAYER_PLAYER ;

	// Init EndGameDBGen Structure
	m_EngineData.EndGameDBGen.bLogOutput	= true ;
	m_EngineData.EndGameDBGen.bOutput		= true ;
	m_EngineData.EndGameDBGen.iOutputMode	= 0 ;
	m_EngineData.EndGameDBGen.iNPiece		= 5 ;
	m_EngineData.EndGameDBGen.iNWhiteMan	= 0 ;
	m_EngineData.EndGameDBGen.iNWhiteKing	= 0 ;
	m_EngineData.EndGameDBGen.iNBlackMan	= 0 ;
	m_EngineData.EndGameDBGen.iNBlackKing	= 0 ;

	// Init Match Structure
	m_EngineData.DXPMatch.bAdjudicate	= true ;
	m_EngineData.DXPMatch.bAlternate	= true ;
	m_EngineData.DXPMatch.bPDN			= true ;

	m_EngineData.DXPMatch.csMatchName	= "DXPMatch" ;

	m_EngineData.DXPMatch.iMaxGames = DXP_DEFAULT_MAXGAMES ;
	m_EngineData.DXPMatch.iMaxMoves = DXP_DEFAULT_MAXMOVES ;
	m_EngineData.DXPMatch.iMaxTime	= DXP_DEFAULT_MAXTIME ;

	m_EngineData.DXPMatch.iNGames		= 0 ;
	m_EngineData.DXPMatch.iNDraw		= 0 ;
	m_EngineData.DXPMatch.iNLoose		= 0 ;
	m_EngineData.DXPMatch.iNWin			= 0 ;
	m_EngineData.DXPMatch.iNUnknown		= 0 ;

	// Init DXP
	m_bDXPGame			= true ;
	m_bDXPMatch			= false ;
	m_bClient			= false ;
	m_bServer			= false ;
	m_bDXPClientTurn	= false ;

	// Init ( clear cPDN Buffer )
	m_pcPDN = NULL ;

	//Reset Game History
	m_EngineData.iGameHistoryActualMove	= 0 ;
	m_EngineData.iGameHistoryMaxMove	= 0 ;

	BoardInit() ;
	
	m_EngineData.GameHistory[0].bTurn = m_EngineData.bTurn ;

	for ( i=0 ; i<5 ; i++ )		// 64Bit
		m_EngineData.GameHistory[0].bbField[i] = m_EngineData.bbField[i] ;

#if _ENGINE_DAMAGE
	if ( bProfile ) {

		m_EngineData.iNormalSearchValue	= ENGINE_PROFILE_DEPTH ;
		m_EngineData.iOpponents			= ENGINE_OPPONENTMODE_ENGINE_ENGINE ;
		m_EngineData.iPPProcessActive	= 1 ;
		m_EngineData.bCWhite			= true ;
		m_EngineData.bCBlack			= true ;

		// SetEngineData(0, ENGINEDATA_STARTSEARCH) ;
	
		if ( m_EngineData.bTurn ) 
			m_EngineData.bComp = m_EngineData.bCWhite ;
		else
			m_EngineData.bComp = m_EngineData.bCBlack ;

		if ( !m_EngineData.bComp && m_EngineData.iEngineMode == ENGINE_MODE_NORMAL && m_EngineData.bBrain )
			m_EngineData.bThink = true ;
		else
			m_EngineData.bThink = false ;
	
		m_pSearch64->CopyBoard(&m_EngineData) ;

		m_pSearch64->Search(0, &m_EngineData ) ;

		exit(0) ;
	}
#endif

	csMessage.Empty() ;		// Message is empty, only Notification ID to EngineApp()
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_NOTIFICATION_ONUPDATE_ENGINEWINDOWTEXT, csMessage) ;

	csMessage.Empty() ;		// Message is empty, only Notification ID to EngineApp()
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_NOTIFICATION_ONUPDATE_ENGINEMODE, csMessage) ;
}


void CEngine::SendEngineMessage(int iMessageSource, int iMessageID, CString csMessage)
{
	int iIndex ;

	m_EngineData.SendEngineMessageLock.Lock() ;

	iIndex = m_EngineData.iOutputStackWrite % ENGINE_OUTPUT_STACK_SIZE ;

	m_EngineData.OutputStack[iIndex].iMessageSource		= iMessageSource ;
	m_EngineData.OutputStack[iIndex].iMessageID			= iMessageID ;
	m_EngineData.OutputStack[iIndex].csOutputMessage	= csMessage ;

	++(m_EngineData.iOutputStackWrite) ;

		m_EngineData.SendEngineMessageLock.Unlock() ;

}


int CEngine::runSearch(int iSearchThreadID)
{
	int iSearchMoves, iThreadTask ;
	CString csMessage, csMove ;


	m_EngineData.bSearch					= true ;
	m_EngineData.bPPSearch[iSearchThreadID]	= true ;
	m_EngineData.bm_idx						= 0 ;	// Was commented out ??????
	m_EngineData.iScoreAnalysisMove			= 0 ;	// Temp. solution, equal score AnalysisMove and BestMove.
	m_EngineData.iIndexBestMove				= 0 ;
	m_EngineData.iScoreBestMove				= 0 ;

//	m_EngineData.SearchThreadData[iSearchThreadID].bLMR =  ( iSearchThreadID == 0 ? true : false ) ;
	m_EngineData.SearchThreadData[iSearchThreadID].bLMR =  m_EngineData.bLMR ;

	if ( m_EngineData.iEngineModeMask & ENGINE_MODEMASK_BOOK_UPDATE ) {	// BOOK_UPDATE has a different Search

	// Future Implementation.....

	} else {
	
		m_EngineData.iPPProcessStatus[iSearchThreadID] = PP_PROCESS_RUN ;

		csMessage.Format("%s", ENGINE_STR_$STATUS) ;
		SendEngineMessage(ENGINE_CONSOLE, ENGINE_$STATUS, csMessage) ;

		iThreadTask = m_EngineData.ThreadData[iSearchThreadID].iThreadTask ;

		if ( iThreadTask == THREAD_TASK_SEARCH ) {														// Search Options

			if ( m_EngineData.iEngineModeMask & ENGINE_MODEMASK_INFINITE_ANALYSIS ) {					// Infinite Analysis
				
				iSearchMoves = m_pSearch64->IASearch(iSearchThreadID, &m_EngineData ) ;

			} else if ( m_EngineData.iEngineModeMask & ENGINE_MODEMASK_ANALYSIS ) {						// Analysis

				iSearchMoves = m_pSearch64->Search(iSearchThreadID, &m_EngineData ) ;				
			
			} else {

				iSearchMoves = m_pSearch64->Search(iSearchThreadID, &m_EngineData ) ;
			}

		} else if ( iThreadTask == THREAD_TASK_MM_SEARCH ) {	// Not implemented....

		} else if ( iThreadTask == THREAD_TASK_SMP_SEARCH ) {

#ifdef _ENGINE_DAMAGE
			if ( iSearchThreadID == 0 )
				iSearchMoves = m_pSearch64->PPSearch(iSearchThreadID, &m_EngineData ) ;
			else
				m_pSearch64->PPSearchSPIdleLoop(iSearchThreadID, &m_EngineData, NULL) ;
#endif
		}

		m_EngineData.iPPProcessStatus[iSearchThreadID]		= PP_PROCESS_IDLE ;
		m_EngineData.iPPSearchThreadStatus[iSearchThreadID] = PP_SEARCHTHREAD_READY ;
		
		csMessage.Format("%s", ENGINE_STR_$STATUS) ;
		SendEngineMessage(ENGINE_CONSOLE, ENGINE_$STATUS, csMessage) ;
	}

	m_EngineData.bSearch					= false ;
	m_EngineData.bPPSearch[iSearchThreadID] = false ;

	return ( iSearchMoves ) ;
}


void CEngine::runPostProcSearch(int iSearchThreadID, int iRunSearch)
{
	int iXMoves ;
	float fAScore, fBScore ;
	CString csMessage, csMove, csAMove, csBMove ;
	PTSTACK pTDepth ;

	// runSearch post-processing
	if ( m_EngineData.iEngineModeMask & ENGINE_MODEMASK_POSITION ) {
		
		if ( m_EngineData.bMove ) {
			
			DecodeEngineMove(iSearchThreadID, m_EngineData.bm_idx) ;
			ExecuteMove(m_EngineData.bm_idx) ;

			if ( m_iCommunicationMode == ENGINE_COMMUNICATIONMODE_GUIDE ) {
				MoveToString(false, &csMove) ;
				csMessage.Format("%s %s", ENGINE_STR_MOVE, csMove) ;

				SendEngineMessage(ENGINE_CONSOLE, ENGINE_MOVE, csMessage) ;
			}

			if ( m_iCommunicationMode == ENGINE_COMMUNICATIONMODE_DXP ) {

				m_EngineData.pTreeStack->bTurn = m_EngineData.bTurn ;
				pTDepth = m_EngineData.SearchThreadData[0].pTreeStack ;

				m_pSearch64->CopyBoard(&m_EngineData) ;						
				iXMoves = m_pSearch64->m_pMoveGenerate64->MoveGenerate(pTDepth ) ;

				if ( iXMoves == 0  ) {											// Check if after Move Opponent can Move, otherwise claim Win (and dont send a Move) !

					m_bDXPGameEnd						= true ;
					m_EngineData.DXPMatch.iGameResult	= DXP_WIN ;

					csMessage.Format("%s", ENGINE_DXP_STR_GAMEEND_WIN) ;		// Win, send GAMEEND
					SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXP_MESSAGE, csMessage) ;
				
				} else if ( iRunSearch == 0 ) {		// No Moves

					m_bDXPGameEnd						= true ;
					m_EngineData.DXPMatch.iGameResult	= DXP_LOSE ;

					csMessage.Format("%s", ENGINE_DXP_STR_GAMEEND_LOOSE) ;		// Loose, send GAMEEND
					SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXP_MESSAGE, csMessage) ;
	
				} else if ( iRunSearch == -1 ) { // DB Value Detected (and bAdjudicate == true )

					if ( m_EngineData.score0 == DATABASE_WIN_VAL ) {

						m_bDXPGameEnd						= true ;
						m_EngineData.DXPMatch.iGameResult	= DXP_WIN ;

						csMessage.Format("%s", ENGINE_DXP_STR_GAMEEND_WIN) ;	// Win, send GAMEEND
						SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXP_MESSAGE, csMessage) ;

					} else if ( m_EngineData.score0 == DATABASE_DRAW_VAL ) {

						m_bDXPGameEnd						= true ;
						m_EngineData.DXPMatch.iGameResult	= DXP_DRAW ;

						csMessage.Format("%s", ENGINE_DXP_STR_GAMEEND_DRAW) ;	// Draw, send GAMEEND
						SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXP_MESSAGE, csMessage) ;

					} else if ( m_EngineData.score0 == DATABASE_LOOSE_VAL ) {

						m_bDXPGameEnd						= true ;
						m_EngineData.DXPMatch.iGameResult	= DXP_LOSE ;

						csMessage.Format("%s", ENGINE_DXP_STR_GAMEEND_LOOSE) ;	// Loose, send GAMEEND
						SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXP_MESSAGE, csMessage) ;
					}

				} else if ( iRunSearch > 0 ) {									// Regular Move

					MoveToDXPString(&csMove) ;

					csMessage.Format("M0000%s", csMove) ;
					SendEngineMessage(ENGINE_CONSOLE, ENGINE_DXPMOVE_MESSAGE, csMessage) ;
				}
			}
		}
	}

	if ( m_EngineData.iEngineModeMask & ENGINE_MODEMASK_ANALYSIS ) {

		if ( m_EngineData.iEngineID == ENGINE_ID_HORIZON3 ) {

			csAMove = m_EngineData.csAnalysisMove ;
			csBMove = m_EngineData.csBestMove ;

		} else {

			DecodeEngineMove(iSearchThreadID, m_EngineData.am_idx) ;
			MoveToString(true, &csAMove) ;

			DecodeEngineMove(iSearchThreadID, m_EngineData.iIndexBestMove) ;
			MoveToString(true, &csBMove) ;
		}

		fAScore		 = (float)m_EngineData.iScoreAnalysisMove ;
		fAScore		/= 100. ; 

		fBScore		 = (float)m_EngineData.iScoreBestMove ;
		fBScore		/= 100. ; 

		csMessage.Format("%s %d %d %s Score %5.2f %s Score %5.2f %s", 
			ENGINE_STR_ANALYSIS, m_EngineData.iScoreAnalysisMove, m_EngineData.iScoreBestMove, csAMove, fAScore, csBMove, fBScore, m_EngineData.csSearchPV  ) ;
	
		SendEngineMessage(ENGINE_CONSOLE, ENGINE_ANALYSIS, csMessage) ;

	}

	if ( m_EngineData.iEngineModeMask & ENGINE_MODEMASK_BOOK_UPDATE ) {
	
		csMessage.Format("%s %d %d",ENGINE_STR_QMRK_BOOKUPDATE, m_iBookUpdate, m_iBookIndex ) ;
		SendEngineMessage(ENGINE_CONSOLE, ENGINE_QMRK_BOOKUPDATE, csMessage) ;

	}
}


void CEngine::runOpeningBookGenerate(void)
{

}


void CEngine::runEndGameDatabasesGenerate(void)
{
}


// !Functions ( ASX_ )
bool CEngine::ASX_DB64(void)
{
	int iASX_DB64_mode ;
	CString csDatabaseData, csToken, csDB64_Mode ;

	csDatabaseData = m_EngineMessage[m_iNMessage].csMessageBody ;

	ParseMessage(&csDatabaseData, &csDB64_Mode) ;

	iASX_DB64_mode = ENGINE_ASX_DB64_COUNT ;

	if ( csDB64_Mode == 'c' )							// Compress
		iASX_DB64_mode = ENGINE_ASX_DB64_COMPRESS ;

	else if ( csDB64_Mode == 'n' )						// Count
		iASX_DB64_mode = ENGINE_ASX_DB64_COUNT ;

	else if ( csDB64_Mode == 'e' )						// Edit
		iASX_DB64_mode = ENGINE_ASX_DB64_EDIT ;

	else if ( csDB64_Mode == 'i' )						// Info
		iASX_DB64_mode = ENGINE_ASX_DB64_INFO ;

	else if ( csDB64_Mode == 'm' )						// Modify
		iASX_DB64_mode = ENGINE_ASX_DB64_MODIFY ;

	else if ( csDB64_Mode == 's' )						// Save
		iASX_DB64_mode = ENGINE_ASX_DB64_SAVE ;

	else if ( csDB64_Mode == 'g' )						// Generate
		iASX_DB64_mode = ENGINE_ASX_DB64_GENERATE ;

	else if ( csDB64_Mode == 't' )						// Test
		iASX_DB64_mode = ENGINE_ASX_DB64_TEST ;

	else if ( csDB64_Mode == 'p' )						// Preload DB Files
		iASX_DB64_mode = ENGINE_ASX_DB64_PRELOAD ;

	else if ( csDB64_Mode == 'z' )						// ZERO ( Clear DB Cache )
		iASX_DB64_mode = ENGINE_ASX_DB64_ZERO ;

	else if ( csDB64_Mode == 'x' )						// Exit
		iASX_DB64_mode = ENGINE_ASX_DB64_EXIT ;

	m_iGenerateSubMode = iASX_DB64_mode ;

	if ( iASX_DB64_mode == ENGINE_ASX_DB64_COMPRESS ) {	// Compress

		m_iDB64Parameter = 0 ;

		ParseMessage(&csDatabaseData, &csToken) ;

		if ( isdigit( csToken[0] ) )
			m_iDB64Parameter = atoi( LPCTSTR(csToken) ) ;

		m_iGenerateMode = ENGINE_GENERATE_MODE_ENDGAMEDATABASES ;

		SetEvent(m_hEventGenerateStart) ;

	}

	if ( iASX_DB64_mode == ENGINE_ASX_DB64_COUNT ) {	// Count

		m_iDB64Parameter = 0 ;

		ParseMessage(&csDatabaseData, &csToken) ;

		if ( isdigit( csToken[0] ) )
			m_iDB64Parameter = atoi( LPCTSTR(csToken) ) ;

		m_iGenerateMode = ENGINE_GENERATE_MODE_ENDGAMEDATABASES ;
		SetEvent(m_hEventGenerateStart) ;
	}

	if (  iASX_DB64_mode == ENGINE_ASX_DB64_MODIFY ) {	// Modify

		m_iGenerateMode = ENGINE_GENERATE_MODE_ENDGAMEDATABASES ;
		SetEvent(m_hEventGenerateStart) ;
	}

	if (  iASX_DB64_mode == ENGINE_ASX_DB64_TEST ) {	// Test

		m_iDB64Parameter = DB64_COMPRESS_DEFAULT ;

		ParseMessage(&csDatabaseData, &csToken) ;

		if ( isdigit( csToken[0] ) )
			m_iDB64Parameter = atoi( LPCTSTR(csToken) ) ;

		m_iGenerateMode = ENGINE_GENERATE_MODE_ENDGAMEDATABASES ;
		SetEvent(m_hEventGenerateStart) ;
	}

	if (  iASX_DB64_mode == ENGINE_ASX_DB64_PRELOAD ) {	// Preload

		m_iDB64Parameter = 0 ;

		ParseMessage(&csDatabaseData, &csToken) ;

		if ( isdigit( csToken[0] ) )
			m_iDB64Parameter = atoi( LPCTSTR(csToken) ) ;

		m_iGenerateMode = ENGINE_GENERATE_MODE_ENDGAMEDATABASES ;
		SetEvent(m_hEventGenerateStart) ;
	}

	if (  iASX_DB64_mode == ENGINE_ASX_DB64_ZERO ) {	// Zero

		m_iGenerateMode = ENGINE_GENERATE_MODE_ENDGAMEDATABASES ;
		SetEvent(m_hEventGenerateStart) ;
	}

	if (  iASX_DB64_mode == ENGINE_ASX_DB64_EXIT ) {	// Exit

		m_iGenerateMode = ENGINE_GENERATE_MODE_ENDGAMEDATABASES ;
		SetEvent(m_hEventGenerateStart) ;
	}

	return ( true ) ;
}


bool CEngine::NMBRSGN_MoveGenerate(void)
{
	int i ;
	PTSTACK pTDepth ;

	pTDepth = m_EngineData.pTreeStack ;

	pTDepth->bTurn = m_EngineData.bTurn ;

	for ( i=0; i<5; i++ )
		pTDepth->bbField[i] = m_EngineData.bbField[i] ;

	m_pSearch64->MoveGenerate(&m_EngineData, true) ;

	return true ;

}


bool CEngine::NMBRSGN_Perft(void)
{
	int i, iDepth, iDepth0, iNSeconds, iNMilliSeconds ; ;
	__int64 iNNodes, iKNodesSec ;
	float fPerftTime ;
	struct timeb PerftStartTime, PerftFinishTime ;
	PTSTACK pTDepth ;
	CString csMessage, csToken ;

	csMessage = m_EngineMessage[m_iNMessage].csMessageBody ;
	
	if ( !ParseMessage(&csMessage, &csToken) )
		return false ;

	iDepth = atoi( LPCTSTR(csToken) ) ;

	if ( !ParseMessage(&csMessage, &csToken) )
		m_Perft_iNMoveGen =  1 ;
	else
		m_Perft_iNMoveGen = atoi( LPCTSTR(csToken) ) ;

	pTDepth = m_EngineData.SearchThreadData[0].pTreeStack ;

	m_pSearch64->CopyBoard(&m_EngineData) ;
	m_pSearch64->MaterialValue(pTDepth) ;


	m_Perft_iCountMoveGen	= 0 ;
	m_Perft_iCountMove		= 0 ;

	for ( i=1; i< TREESTACK_SIZE ; i++ ) {
		if ( i % 2 )
			(pTDepth+i)->bTurn = !( pTDepth->bTurn ) ;
		else
			(pTDepth+i)->bTurn = pTDepth->bTurn ;
	}

	if ( iDepth > 0 )
		iDepth0 = 1 ;
	else {
		iDepth	= abs(iDepth) ;
		iDepth0 = iDepth ;
	}

	for ( i=iDepth0; i<=iDepth; i++ ) {

		// Timer Start
		ftime( &PerftStartTime ) ;

		iNNodes = Perft(i, pTDepth) ;

		// Timer Finish
		ftime( &PerftFinishTime ) ;

		iNSeconds		= (int)( PerftFinishTime.time		- PerftStartTime.time ) ;
		iNMilliSeconds	= (int)( PerftFinishTime.millitm	- PerftStartTime.millitm ) ;

		fPerftTime  = (float)( ( 1000 * iNSeconds )  + iNMilliSeconds ) ;
		fPerftTime	/= 1000. ;

		iKNodesSec = (__int64)( iNNodes / ( 1000 * fPerftTime ) ) ;

		csMessage.Format("Perft(%d)\tN = %I64d\t%7.2f sec.\tKN/sec = %d", i, iNNodes, fPerftTime, iKNodesSec) ;
		SendEngineMessage(ENGINE_CONSOLE, ENGINE_NMBRSGN_PERFT, csMessage) ;

	}

	return ( true ) ;
}


bool CEngine::NMBRSGN_Time(void)
{
	int i, iCount, iNSeconds, iNMilliSeconds ;
	__int64 iXCount ;
	float fTime ;
	struct timeb StartTime, FinishTime ;
	CString csMessage, csToken ;
	PTSTACK pTDepth ;

	csMessage = m_EngineMessage[m_iNMessage].csMessageBody ;
	
	if ( !ParseMessage(&csMessage, &csToken) )
		return false ;

	iCount = atoi( LPCTSTR(csToken) ) ;

	m_pSearch64->CopyBoard(&m_EngineData) ;

	iXCount = 0 ;
	pTDepth = m_EngineData.SearchThreadData[0].pTreeStack ;

	// Timer Start
	ftime( &StartTime ) ;

	for ( i=0; i<iCount; i++ )
		 iXCount += m_pSearch64->m_pMoveGenerate64->MoveGenerate(pTDepth ) ;;

	// Timer Finish
	ftime( &FinishTime ) ;

	iNSeconds		= (int)( FinishTime.time - StartTime.time ) ;
	iNMilliSeconds	= (int)( FinishTime.millitm	- StartTime.millitm ) ;

	fTime  = (float)( ( 1000 * iNSeconds )  + iNMilliSeconds ) ;
	fTime	/= 1000. ;

	csMessage.Format("MoveGenerate = %7.2f sec.", fTime) ;
	SendEngineMessage(ENGINE_CONSOLE, ENGINE_NMBRSGN_TIME, csMessage) ;

	return ( true ) ;
}


bool CEngine::At_Run(void)
{
	int iNCommand ;
	CString csFileName, csMessageBody, csCommand ;
	CFile CFile_Script ;

	csMessageBody = m_EngineMessage[m_iNMessage].csMessageBody ;
	ParseMessage(&csMessageBody, &csFileName) ;

	if ( csFileName.IsEmpty() )
		csFileName.Format("script.dat") ;

	csFileName = m_csScriptsPath + csFileName ; 

	m_Script.bScript			= true ;
	m_Script.csScriptFileName	= csFileName ;
	m_Script.iCommand			= 0 ;
	m_Script.iNCommand			= 0 ;

	if ( !CFile_Script.Open(csFileName, CFile::modeRead ) )
		return ( false ) ;

	CArchive CArchive_Script(&CFile_Script, CArchive::load ) ;

	iNCommand = 0 ;

	while ( CArchive_Script.ReadString(csCommand) ) {

		m_Script.csCommand[iNCommand] = csCommand ;
		++iNCommand ;
	}

	m_Script.iNCommand = iNCommand ;

	SetEvent(m_hEventScriptStart) ;

	return ( true ) ;
}


bool CEngine::At_RunNext(void)
{

	return ( true ) ;
}
