// CENGINE.H

#if !defined ( _CENGINE_H_ )

#define _CENGINE_H_

#define ENGINENAME_D					"Damage 12.2"
#define ENGINEAUTHOR_D					"Bert Tuyt"

#define ENGINENAME_H4					"Horizon 4.1"
#define ENGINEAUTHOR_H4					"Feike Boomstra & Bert Tuyt"

#define ENGINE_EVAL64_TESTFILE			"EvalTestPos.txt"
#define ENGINE_EVAL64_OUTPUTTESTFILE	"EvalOuputTestPos.txt"

#define ENGINE_MODE_NORMAL				0
#define ENGINE_MODE_ANALYSIS			1

#define ENGINE_GENERATE_MODE_NONE				0
#define ENGINE_GENERATE_MODE_OPENINGBOOK		1
#define ENGINE_GENERATE_MODE_BREAKTHROUGHTABLE	2
#define ENGINE_GENERATE_MODE_ENDGAMEDATABASES	3

#define ENGINE_PROFILE_DEPTH			15

#define ENGINE_OPTION_OPENINGBOOK		0
#define ENGINE_OPTION_ENDGAMEDATABASE	1
#define ENGINE_OPTION_PERMANENTBRAIN	2
#define ENGINE_OPTION_HASHTABLE			3
#define ENGINE_OPTION_DBCACHESIZE		4
#define ENGINE_OPTION_DBMAXPIECE		5

#define ENGINEDATA_STARTSEARCH			0
#define ENGINEDATA_STOPSEARCH			1

#define ENGINE_ASX_DB64_COMPRESS		0
#define ENGINE_ASX_DB64_COUNT			1
#define ENGINE_ASX_DB64_EDIT			2
#define ENGINE_ASX_DB64_INFO			3
#define ENGINE_ASX_DB64_MODIFY			4
#define ENGINE_ASX_DB64_SAVE			5
#define ENGINE_ASX_DB64_GENERATE		6
#define ENGINE_ASX_DB64_TEST			7
#define ENGINE_ASX_DB64_PRELOAD			8
#define ENGINE_ASX_DB64_ZERO			9
#define ENGINE_ASX_DB64_EXIT			10


class CSearch64 ;
class CPDN ;
class CGuide ;

class CEngine
{

public:

	CEngine() ;
	~CEngine() ;
	
	HANDLE m_hEventBookStart ;
	HANDLE m_hEventGenerateStart ;
	HANDLE m_hEventScriptStart ;
	HANDLE m_hEventGameStart ;
	HANDLE m_hEventGameStop ;
	HANDLE m_hEventMatchStart ;

	CMutex			m_SendEngineMessageLock ;

	bool			m_bDebug ;
	bool			m_bGuide ;
	bool			m_bDXPGame ;
	bool			m_bDXPMatch ;
	bool			m_bClient ;
	bool			m_bServer ;
	bool			m_bDXPClientTurn ;
	bool			m_bDXPGameEnd ;

	char*			m_pcPDN ;
	int				m_iPDNMoves ;
	PGAMEHEADER		m_pPDNGameHeader ;

	int				m_Perft_iNMoveGen ;
	__int64			m_Perft_iCountMoveGen ;
	__int64			m_Perft_iCountMove ;

	int				m_iCommunicationMode ;
	int				m_iGenerateMode ;
	int				m_iGenerateSubMode ;
	int				m_iDB64Parameter ;
	int				m_iData[51] ;
	int				m_iEngineStatus ;
	int				m_iNMessage ;

	int				m_iBookIndex ;
	int				m_iBookUpdate ;

	CString			m_csApplicationPath ;
	CString			m_csBookPath ;
	CString			m_csDatabasePath ;
	CString			m_csGamesPath ;
	CString			m_csScriptsPath ;
	CString			m_csTablesPath ;
	
	ENGINEDATA		m_EngineData ;
	ENGINEMESSAGE	m_EngineMessage[ENGINE_MAX_MESSAGE] ;

	SCRIPT			m_Script ;

	CSearch64*				m_pSearch64 ;
	CPDN*					m_pPDN ;	// Pointer to CPDN Class
	CGuide*					m_pGuide ;	// Pointer to CGuide Class

	__int64	Perft(int, PTSTACK);

	void	BoardLog(int*, bool, CString*) ;
	bool	DamageLog(int) ;
	float	DamageTime(int) ;
	void	FormatDamageTime(float, CString*) ;
	void	ClearOutput(int) ;
	void	EvalOutput(void) ;

	bool	FENRead(char*) ;
	bool	PDNCopyBoard(void) ;
	bool	PDNCopyMoves(void) ;

	void	MoveToDXPString(CString*) ;
	void	MoveToString(bool, CString*) ;

	void	MoveListToString(CString*) ;
	void	PositionToString(CString*) ;

	bool	DXPBallot(int, CString*, CString*, CString*) ;
	bool	DXPBallotGame(int) ;
	bool	DXPUpdateStat(int) ;
	void	DXPGameFormat(CString*) ;
	void	DXPStat(void) ;
	bool	DXPStatMatchFile(CString*) ;

	bool	DXPPositionToString(CString*) ;
	bool	DXPStringToPosition(bool, CString*) ;
	bool	StringToPosition(bool, CString*) ;

	void	DXPClientStart(void) ;
	int		DecodeDXPMessage(CString) ;
	int		DecodeMessage(CString) ;

	bool	CommandMessage(CString, CString*, CString*) ;
	bool	ParseMessage(CString*, CString* ) ;
	bool	ParseGameMove(CString*, int* ) ;

	void	BoardClear(void) ;
	void	BoardInit(void) ;
	bool	BoardUpdate() ;

	bool	DecodeBlitzGameInfo(void) ;
	void	DecodeEngineMove(int, int) ;

	int		DecodeDXPMove(void) ;
	int		DecodeGUIMove(void) ;
	void	ProcessMove(int*, int*) ;
	void	FormatMove(int*, CString*) ;

	bool	EngineMode(void) ;

	void	ExecuteDXPMessage(void) ;
	void	ExecuteMessage(void) ;

	void	ExecuteMove(int) ;
	void	Forward(void) ;

	bool	Game(void) ;

	void	DXPGo(void) ;
	void	MatchGo(void) ;
	void	Go(void) ;

	void	Halt(void) ;
	void	Halt(int) ;
	void	InitEngine(bool) ;

	void	JumpTo(void) ;
	bool	JumpTo(int) ;

	void	LevelsBlitzGame(void) ;
	void	LevelsLongGame(void) ;
	void	LevelsFixedDepth(void) ;
	void	LevelsFixedTime(void) ;
	void	LevelsInfiniteAnalysis(void) ;

	void	DXPMove(void) ;
	void	Move(void) ;

	void	New(void) ;

	void	OfferDraw(void) ;
	void	Opponents(void) ;
	void	Ping(void) ;
	void	Pong(CString) ;
	bool	Position(void) ;
	void	RequestGame(void) ;
	void	RequestPosition(void) ;
	void	SetComment(void) ;
	bool	SetOption(void) ;
	void	ShowWindow(void) ;

	void	OnOptions(void) ;

	// PP (Parallel Processing) Routines
	bool	PPMode(bool, int) ;
	bool	PPProc(void) ;

	void	SearchThreadStart(int) ;

	void	ThreadStart(void) ;
	void	ThreadStop(bool, int, bool) ;
	
	void	SendEngineMessage(int, int, CString) ;

	int		runSearch(int) ;
	void	runPostProcSearch(int, int) ;

	void	runOpeningBookGenerate(void) ;
	void	runEndGameDatabasesGenerate(void) ;

	bool	ASX_DB64(void) ;

	bool	At_Run(void) ;
	bool	At_RunNext(void) ;

	bool	NMBRSGN_MoveGenerate(void) ;
	bool	NMBRSGN_Perft(void) ;
	bool	NMBRSGN_Time(void) ;

	// $ ( DLR_ ) Routines
	bool	DLR_ClearCount(void) ;
	bool	DLR_ClearOutput(void) ;
	bool	DLR_Debug(void) ;
	void	DLR_DXP(void) ;
	bool	DLR_DXP_MatchMode(void) ;
	void	DLR_DXP_Prm_Display(void) ;
	bool	DLR_Eval(void) ;
	bool	DLR_Info(void) ;
	bool	DLR_Game(void) ;
	bool	DLR_Log(void) ;
	bool	DLR_Match(void) ;
	bool	DLR_PDN(void) ;
	bool	DLR_Prm(void) ;
	void	DLR_Prm_Display(void) ;
	void	DLR_Info_PVPosition(void) ;

private:

	int	m_iMoveData[23] ;
	
} ;

#endif	// _CENGINE_H_